"""
This contains the basic residue template and residue building libraries
typically used in modelling applications
"""

from collections import OrderedDict, defaultdict
import copy as _copy
import numpy as np
import os
from parmed.residue import AminoAcidResidue, RNAResidue, DNAResidue
from parmed.structure import Structure
from parmed.topologyobjects import Atom, Bond, AtomList, TrackedList
from parmed.utils.six import iteritems
from parmed.exceptions import IncompatiblePatchError, MoleculeError
import warnings

__all__ = ['PROTEIN', 'NUCLEIC', 'SOLVENT', 'UNKNOWN', 'ResidueTemplate',
           'ResidueTemplateContainer', 'PatchTemplate']

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class _ResidueType(object):
    """ Singleton for various types of residues """
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return '<ResidueType %s>' % self.name

    def __str__(self):
        return self.name

PROTEIN = _ResidueType('PROTEIN')
NUCLEIC = _ResidueType('NUCLEIC')
SOLVENT = _ResidueType('SOLVENT')
UNKNOWN = _ResidueType('UNKNOWN')

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class ResidueTemplate(object):
    """
    This is a residue template, which contains a listing of the atoms in the
    residue template as well as a mapping of which atoms are bonded to other
    atoms.

    Parameters
    ----------
    name : str, optional
        If provided, this is the name of the residue

    Attributes
    ----------
    atoms : :class:`AtomList`
        List of atoms in this residue
    bonds : :class:`TrackedList`
        List of the bonds between the atoms in this residue
    coordinates : np.ndarray(natom, 3)
        The partial atomic coordinates
    connections : list of :class:`Atom`
        A list of all atoms that should form connections with atoms of another
        residue *besides* the head and tail atoms
    head : :class:`Atom` or None
        The atom that is connected to the residue that comes before this one
    tail : :class:`Atom` or None
        The atom that is connected to the *next* residue after this one
    first_patch : :class:`ResidueTemplate` or None
        If it is not None, this is the patch whose tail is added to the head
        atom of this residue when this residue is the first in a chain
    last_patch : :class:`ResidueTemplate` or None
        If it is not None, this is the patch whose head is added to the tail
        atom of this residue when this residue is the last in a chain
    groups : list of list(:class:`Atom`)
        If set, each group is a list of Atom instances making up each group
    override_level : integer
        For use with OpenMM ResidueTemplates. If OpenMM ForceField is given multiple
        identically-matching residue templates with the same names it choses
        (overrides with) the one with the highest override_level
        (overrideLevel in OpenMM). Default is 0.
    """

    def __init__(self, name=''):
        self.atoms = AtomList()
        self.bonds = TrackedList()
        self.lonepairs = list() # TODO: Should this be a TrackedList?
        self.anisotropies = list()
        self.name = name
        self.head = None
        self.tail = None
        self.connections = []
        self.type = UNKNOWN
        self.first_patch = None
        self.last_patch = None
        self.groups = []
        self.override_level = 0
        self._map = dict()
        self._impr = []

    def __repr__(self):
        if self.head is not None:
            head = self.head.name
        else:
            head = 'None'
        if self.tail is not None:
            tail = self.tail.name
        else:
            tail = 'None'
        return '<%s %s: %d atoms; %d bonds; head=%s; tail=%s>' % (
                    type(self).__name__, self.name, len(self.atoms),
                    len(self.bonds), head, tail)

    @property
    def map(self):
        return self._map

    def add_atom(self, atom):
        """ Adds an atom to this residue template

        Parameters
        ----------
        atom : :class:`Atom`
            The atom to add to this residue

        Raises
        ------
        ValueError if ``atom`` has the same name as another atom in this
        residue already
        """
        if atom.name in self._map:
            raise ValueError('Residue already has atom named %s' % atom.name)
        atom.residue = self
        self.atoms.append(atom)
        self._map[atom.name] = atom

    def delete_atom(self, atom):
        """ Delete an atom from this residue template, along with corresponding bonds.

        Parameters
        ----------
        atom : :class:`Atom` or str
            The atom or atom name to be deleted

        """
        if type(atom) is str:
            atom_name = atom
        else:
            atom_name = atom.name

        if atom_name not in self._map:
            raise KeyError("Could not find atom '%s' in ResidueTemplate, which contains atoms: %s" % (atom_name, list(self._map.keys())))

        atom = self._map[atom_name]

        # Adjust head and tail if needed
        if self.head == atom:
            self.head = None
        if self.tail == atom:
            self.tail = None

        # Remove all bonds involving this atom
        for bond in list(self.bonds):
            if (bond.atom1 == atom) or (bond.atom2 == atom):
                self.delete_bond(bond)

        # Remove all impropers involving this atom
        for impr in list(self._impr):
            if atom in impr:
                self._impr.remove(impr)

        # Disconnect the atom from the residue so that it does not trigger atom.residue.delete_atom(atom)
        atom.residue = None

        # Remove the atom from the ResidueTemplate
        del self._map[atom_name]
        self.atoms.remove(atom)

    def add_bond(self, atom1, atom2, order=1.0):
        """ Adds a bond between the two provided atoms in the residue

        Parameters
        ----------
        atom1 : :class:`Atom` or int or str
            One of the atoms in the bond. It must be in the ``atoms`` list of
            this ResidueTemplate. It can also be the atom index (index from 0)
            of the atom in the bond.
        atom2 : :class:`Atom` or int or str
            The other atom in the bond. It must be in the ``atoms`` list of this
            ResidueTemplate. It can also be the atom index (index from 0) of the
            atom in the bond.
        order : float
            The bond order of this bond. Bonds are classified as follows:
                1.0 -- single bond
                2.0 -- double bond
                3.0 -- triple bond
                1.5 -- aromatic bond
                1.25 -- amide bond
            Default is 1.0

        Raises
        ------
        IndexError if atom1 or atom2 are integers that are out of range of the
        number of atoms already in this template

        RuntimeError if atom1 or atom2 are :class:`Atom` instances but they are
        *not* in the atoms list of this ResidueTemplate

        Notes
        -----
        If atom1 and atom2 are already bonded, this routine does nothing. If
        atom1 or atom2 are strings, then they will match the first instance of
        the atom name that is the same as the atom name passed.
        """
        if not isinstance(atom1, Atom):
            atom1 = self[atom1]
        if not isinstance(atom2, Atom):
            atom2 = self[atom2]
        if atom1.list is not self.atoms or atom2.list is not self.atoms:
            raise RuntimeError('Both atoms must belong to template.atoms')
        # Do not add the same bond twice
        if atom1 not in atom2.bond_partners:
            self.bonds.append(Bond(atom1, atom2, order=order))

    def delete_bond(self, bond):
        """ Delete a bond from this residue template.

        Parameters
        ----------
        bond : :class:`Bond`
            The bond to be deleted

        """
        if bond in self.bonds:
            bond.delete()
            self.bonds.remove(bond)
        else:
            raise ValueError('The specified bond {} does not belong to this residue {}'.format(bond, self))

    @classmethod
    def from_residue(cls, residue):
        """
        This constructor creates a ResidueTemplate from a particular Residue
        object
        Parameters
        ----------
        residue : :class:`Residue`
            The residue from which to create a template
        """
        inst = cls(name=residue.name)
        for atom in residue:
            inst.add_atom(_copy.copy(atom))
        for atom in residue:
            for bond in atom.bonds:
                try:
                    i1 = residue.atoms.index(bond.atom1)
                    i2 = residue.atoms.index(bond.atom2)
                except ValueError:
                    if bond.atom1 in residue:
                        oatom = bond.atom2
                        idx = residue.atoms.index(bond.atom1)
                    else:
                        oatom = bond.atom1
                        idx = residue.atoms.index(bond.atom2)
                    if oatom.residue.idx == residue.idx - 1:
                        inst.head = inst.atoms[idx]
                    elif oatom.residue.idx == residue.idx + 1:
                        inst.tail = inst.atoms[idx]
                    elif oatom.residue.idx == residue.idx:
                        # Don't know WHAT to do with it
                        warnings.warn('Cannot determine head/tail for '
                                      'unordered residues.')
                    else:
                        # Disulfide or something... not head or tail
                        inst.connections.append(inst.atoms[idx])
                else:
                    inst.add_bond(i1, i2)
        return inst

    @property
    def coordinates(self):
        """ Atomic coordinates, in Angstroms, of all atoms in the template """
        try:
            return self._crd
        except AttributeError:
            self._crd = np.array([[a.xx, a.xy, a.xz] for a in self])
        return self._crd

    @property
    def empirical_chemical_formula(self):
        """ Return the empirical chemical formula (in Hill notation) as a string (e.g. 'H2O', 'C6H12'), omitting EPs """
        # Count number of appearances of each element
        element_count = defaultdict(int)
        for atom in self.atoms:
            element_count[atom.element_name] += 1
        # Pop EPs if they are present, since they are not chemical
        if 'EP' in element_count:
            element_count.pop('EP')
        # Render to string using Hill notation
        # https://en.wikipedia.org/wiki/Chemical_formula#Hill_system
        def format_and_pop_element(element_count, element):
            count = element_count.pop(element)
            if count == 1:
                return element
            return element + str(count)

        chemical_formula = ''
        # If carbon is present, first list C, then H (if present)
        if 'C' in element_count:
            chemical_formula += format_and_pop_element(element_count, 'C')
            if 'H' in element_count:
                chemical_formula += format_and_pop_element(element_count, 'H')
        # Remaining elements are listed alphabetically
        alphabetical_elements = sorted(element_count.keys())
        for element in alphabetical_elements:
            chemical_formula += format_and_pop_element(element_count, element)

        return chemical_formula

    @property
    def net_charge(self):
        return sum([a.charge for a in self])

    # Make ResidueTemplate look like a container of atoms, also indexable by the
    # atom name
    def __len__(self):
        return len(self.atoms)

    def __iter__(self):
        return iter(self.atoms)

    def __contains__(self, atom):
        if isinstance(atom, Atom):
            return atom in self.atoms
        if isinstance(atom, str):
            return atom in self._map
        raise AssertionError('Should not be here!')

    def __copy__(self):
        other = type(self)(name=self.name)

        for atom in self.atoms:
            other.add_atom(_copy.copy(atom))
        for bond in self.bonds:
            other.add_bond(bond.atom1.idx, bond.atom2.idx)
        other.type = self.type

        if self.head is not None:
            other.head = other.atoms[self.head.idx]
        if self.tail is not None:
            other.tail = other.atoms[self.tail.idx]
        for connection in self.connections:
            other.connections.append(other.atoms[connection.idx])
        other.first_patch = self.first_patch
        other.last_patch = self.last_patch

        return other

    def __getitem__(self, idx):
        if isinstance(idx, str):
            for atom in self.atoms:
                if atom.name == idx:
                    return atom
            raise IndexError('Atom %s not found in %s' % (idx, self.name))
        elif isinstance(idx, (list, tuple)):
            return [self[key] for key in  idx]
        else:
            return self.atoms[idx]

    def fix_charges(self, to=None, precision=4):
        """
        Adjusts the partial charge of all atoms in the residue to match the
        requested target charge. The default target charge is the closest
        integer

        Parameters
        ----------
        to : float, optional
            The desired net charge of this residue template. Default is the
            closest integer charge
        precision : int, optional
            The number of decimal places that each charge should be rounded to.
            Default is 4

        Returns
        -------
        self : :class:`ResidueTemplate`
            The current residue template whose charges are being modified

        Notes
        -----
        This method modifies the atomic charges of this residue template
        in-place. Any residual charge (which is accumulated roundoff beyond the
        requested precision) is added to the first atom of the residue. This
        will typically be 10^-precision in magnitude, and should almost never be
        higher than 2*10^-precision. As long as a reasonable precision is chosen
        (no fewer than 3 or 4 decimal places), this will have only a negligible
        impact on a force field.

        If provided, "to" will be rounded to the ``precision``'th decimal place
        to make sure that the sum of the charges come out as close as possible
        to the target charge while still obeying the requested precision.

        Raises
        ------
        ValueError
            If you try to call fix_charges on a residue template with no atoms
        """
        if not self.atoms:
            raise ValueError('Cannot fix charges on an empty residue')
        net_charge = self.net_charge
        if to is None:
            to = round(net_charge)
        else:
            # We need to make sure
            to = round(to, precision)
        if net_charge == to:
            return self

        smear = (to - net_charge) / len(self)
        for atom in self:
            atom.charge = round(atom.charge + smear, precision)

        # Dump the extra tiny bit (O(10^-precision)) on the first atom
        self.atoms[0].charge += to - sum(atom.charge for atom in self.atoms)

        return self

    def apply_patch(self, patch, precision=4):
        """
        Apply the specified PatchTemplate to the ResidueTemplate.

        This only handles patches that affect a single residue.

        An exception is thrown if patch is incompatible because
        * The patch specifies that an atom is to be deleted that doesn't exist in the residue
        * A bond specified as being added in the patch does not have both atom names present after adding/deleting atoms from the patch
        * The new net charge is not integral to the specified precision
        * The residue is not modified in any way (no atoms or bonds added/changed/deleted)

        Parameters
        ----------

        patch : PatchTemplate
            The patch to apply to this residue

        precision : int, optional
            Each valid patch should be produce a net charge that is integral to
            this many decimal places.
            Default is 4

        Returns
        -------

        residue : ResidueTemplate
            A new ResidueTemplate corresponding to the patched residue is returned.
            The original remains unmodified.

        """
        # Create a copy
        # TODO: Once ResidueTemplate.from_residue() actually copies all info, use that instead?
        residue = _copy.copy(self)
        # Record whether we've actually modified the residue.
        modifications_made = False
        # Delete atoms
        for atom_name in patch.delete_atoms:
            try:
                residue.delete_atom(atom_name)
                modifications_made = True
            except (KeyError, MoleculeError) as e:
                if atom_name.startswith('D') and atom_name[1:] in self and atom_name[1:] in patch.delete_atoms:
                    # This is a Drude particle.  We're also deleting its parent atom, so don't report an error.
                    pass
                else:
                    raise IncompatiblePatchError('Atom %s could not be deleted from the patched residue: atoms are %s (exception: %s)' % (atom_name, list(residue._map.keys()), str(e)))
        # Add or replace atoms
        for atom in patch.atoms:
            if atom.name in residue:
                # Overwrite type and charge
                residue[atom.name].type = atom.type
                residue[atom.name].charge = atom.charge
            else:
                residue.add_atom(Atom(name=atom.name, type=atom.type, charge=atom.charge))
            modifications_made = True
        # Add bonds
        for (atom1_name, atom2_name, order) in patch.add_bonds:
            try:
                # Remove dangling bonds
                for name in [atom1_name, atom2_name]:
                    if residue.head and (name == residue.head.name):
                        residue.head = None
                    if residue.tail and (name == residue.tail.name):
                        residue.tail = None
                # Add bond
                residue.add_bond(atom1_name, atom2_name, order)
                modifications_made = True
            except (IndexError, MoleculeError) as e:
                raise IncompatiblePatchError('Bond %s-%s could not be added to patched residue: atoms are %s (exception: %s)' % (atom1_name, atom2_name, list(residue._map.keys()), str(e)))
        # Delete impropers
        for impr in patch.delete_impropers:
            try:
                residue._impr.remove(impr)
                # removal of impropers doesn't do anything as far as OpenMM is concerned, so don't note this as a modification having been made
            except ValueError as e:
                raise IncompatiblePatchError('Improper %s was not found in residue to be patched.' % impr)
        # Check that the net charge is integral.
        net_charge = residue.net_charge
        is_integral = (round(net_charge, precision) - round(net_charge)) == 0.0
        if not is_integral:
            raise IncompatiblePatchError('Patch is not compatible with residue due to non-integral charge (charge was %f).' % net_charge)
        # Ensure residue is connected
        import networkx as nx
        G = residue.to_networkx(False)
        if not nx.is_connected(G):
            components = [ c for c in nx.connected_components(G) ]
            raise IncompatiblePatchError('Patched residue bond graph is not a connected graph: %s' % str(components))
        # Make sure the patch has actually modified the residue
        if not modifications_made:
            raise IncompatiblePatchError('Patch did not modify residue.')

        return residue

    def patch_is_compatible(self, patch):
        """Determine whether a specified patch is compatible with this residue.

        Compatibility is determined by whether Residue.Template.apply_patch(patch) raises as
        exception or not.

        Parameters
        ----------
        patch : PatchTemplate
            The patch to be applied to this residue.

        Returns
        -------
        is_compatible : bool
            True if patch is compatible with the residue; False if not.

        """
        try:
            self.apply_patch(patch)
            return True
        except IncompatiblePatchError:
            return False

    def to_networkx(self, include_extra_particles=True):
        """ Create a NetworkX graph of atoms and bonds

        Parameters
        ----------
        include_extra_particles : bool
            Whether to include "atoms" that actually represent extra particles (atomic_number == 0).

        Returns
        -------
        G : :class:`networkx.Graph`
            A NetworkX Graph representing the molecule

        """
        import networkx
        G = networkx.Graph()
        for atom in self.atoms:
            if atom.atomic_number != 0 or include_extra_particles:
                G.add_node(atom.name, charge=atom.charge, type=atom.type)
        for bond in self.bonds:
            if (bond.atom1.atomic_number != 0 and bond.atom2.atomic_number != 0) or include_extra_particles:
                G.add_edge(bond.atom1.name, bond.atom2.name)
        return G

    def to_dataframe(self):
        """ Create a pandas dataframe from the atom information

        Returns
        -------
        df : :class:`pandas.DataFrame`
            The pandas DataFrame with all of the atomic properties

        Notes
        -----
        The DataFrame will be over all atoms. The columns will be the attributes
        of the atom (as well as its containing residue). Some columns will
        *always* exist. Others will only exist if those attributes have been set
        on the Atom instances (see the :class:`Atom` docs for possible
        attributes and their meaning). The columns that will always be present
        are:

            - number : int
            - name : str
            - type : str
            - atomic_number : int
            - charge : float
            - mass : float
            - nb_idx : int
            - solvent_radius : float
            - screen : float
            - occupancy : float
            - bfactor : float
            - altloc : str
            - tree : str
            - join : int
            - irotat : int
            - rmin : float
            - epsilon : float
            - rmin_14 : float
            - epsilon_14 : float

        The following attributes are optionally present if they were present in
        the original file defining the structure:

            - xx : float (x-coordinate position)
            - xy : float (y-coordinate position)
            - xz : float (z-coordinate position)
            - vx : float (x-coordinate velocity)
            - vy : float (y-coordinate velocity)
            - vz : float (z-coordinate velocity)
        """
        import pandas as pd
        ret = pd.DataFrame()

        ret['number'] = [atom.number for atom in self.atoms]
        ret['name'] = [atom.name for atom in self.atoms]
        ret['type'] = [atom.type for atom in self.atoms]
        ret['atomic_number'] = [atom.atomic_number for atom in self.atoms]
        ret['charge'] = [atom.charge for atom in self.atoms]
        ret['mass'] = [atom.mass for atom in self.atoms]
        ret['nb_idx'] = [atom.nb_idx for atom in self.atoms]
        ret['solvent_radius'] = [atom.solvent_radius for atom in self.atoms]
        ret['screen'] = [atom.screen for atom in self.atoms]
        ret['occupancy'] = [atom.occupancy for atom in self.atoms]
        ret['bfactor'] = [atom.bfactor for atom in self.atoms]
        ret['altloc'] = [atom.altloc for atom in self.atoms]
        ret['tree'] = [atom.tree for atom in self.atoms]
        ret['join'] = [atom.join for atom in self.atoms]
        ret['irotat'] = [atom.irotat for atom in self.atoms]
        ret['rmin'] = [atom.rmin for atom in self.atoms]
        ret['epsilon'] = [atom.epsilon for atom in self.atoms]
        ret['rmin_14'] = [atom.rmin_14 for atom in self.atoms]
        ret['epsilon_14'] = [atom.epsilon_14 for atom in self.atoms]
        ret['resname'] = [atom.residue.name for atom in self.atoms]

        # Now for optional attributes
        # Coordinates
        try:
            coords = pd.DataFrame(
                    [[atom.xx, atom.xy, atom.xz] for atom in self.atoms],
                    columns=['xx', 'xy', 'xz']
            )
        except AttributeError:
            pass
        else:
            ret = ret.join(coords)
        # Velocities
        try:
            vels = pd.DataFrame(
                    [[atom.vx, atom.vy, atom.vz] for atom in self.atoms],
                    columns=['vx', 'vy', 'vz']
            )
        except AttributeError:
            pass
        else:
            ret = ret.join(vels)
        return ret

    def to_structure(self):
        """
        Generates a Structure instance with a single residue from this
        ResidueTemplate

        Returns
        -------
        struct : :class:`parmed.structure.Structure`
            The Structure with all of the bonds and connectivity of this
            template
        """
        struct = Structure()
        for atom in self:
            struct.add_atom(_copy.copy(atom), self.name, 0)
        for bond in self.bonds:
            struct.bonds.append(Bond(struct.atoms[bond.atom1.idx],
                                     struct.atoms[bond.atom2.idx])
            )
        return struct

    def save(self, fname, format=None, overwrite=False, **kwargs):
        """
        Saves the current ResidueTemplate in the requested file format.
        Supported formats can be specified explicitly or determined by file-name
        extension. The following formats are supported, with the recognized
        suffix shown in parentheses:

            - MOL2 (.mol2)
            - MOL3 (.mol3)
            - OFF (.lib/.off)
            - PDB (.pdb)
            - PQR (.pqr)

        Parameters
        ----------
        fname : str
            Name of the file to save. If ``format`` is ``None`` (see below), the
            file type will be determined based on the filename extension. If the
            type cannot be determined, a ValueError is raised.
        format : str, optional
            The case-insensitive keyword specifying what type of file ``fname``
            should be saved as. If ``None`` (default), the file type will be
            determined from filename extension of ``fname``
        overwrite : bool, optional
            If True, allow the target file to be overwritten. Otherwise, an
            IOError is raised if the file exists. Default is False
        kwargs : keyword-arguments
            Remaining arguments are passed on to the file writing routines that
            are called by this function

        Raises
        ------
        ValueError if either filename extension or ``format`` are not recognized
        TypeError if the structure cannot be converted to the desired format for
        whatever reason
        """
        from parmed.amber.offlib import AmberOFFLibrary
        from parmed.formats.mol2 import Mol2File
        extmap = {
                '.mol2' : 'MOL2',
                '.mol3' : 'MOL3',
                '.off' : 'OFFLIB',
                '.lib' : 'OFFLIB',
                '.pdb' : 'PDB',
                '.pqr' : 'PQR',
        }
        if format is not None:
            format = format.upper()
        else:
            base, ext = os.path.splitext(fname)
            if ext in ('.bz2', '.gz'):
                ext = os.path.splitext(base)[1]
            if ext in extmap:
                format = extmap[ext]
            else:
                raise ValueError('Could not determine file type of %s' % fname)
        if format == 'MOL2':
            Mol2File.write(self, fname, mol3=False, **kwargs)
        elif format == 'MOL3':
            Mol2File.write(self, fname, mol3=True, **kwargs)
        elif format in ('OFFLIB', 'OFF'):
            AmberOFFLibrary.write({self.name : self}, fname, **kwargs)
        elif format in ('PDB', 'PQR'):
            self.to_structure().save(fname, format=format, overwrite=overwrite,
                                     **kwargs)
        else:
            raise ValueError('Unrecognized format for ResidueTemplate save')

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class PatchTemplate(ResidueTemplate):
    """
    A residue patch (typically used for CHARMM) that is used to modify existing
    residues in some way (e.g., terminal patches, disulfide bridges, etc.)

    Parameters
    ----------
    name : str, optional
        If provided, this is the name of the residue

    Attributes
    ----------
    add_bonds : list of (str, str, order)
        List of bonds that need to be added in applying the patch
    delete_atoms : list of str
        List of atom names that need to be deleted in applying the patch
    delete_impropers : list of tuple of str
        List of impropers (tuple of atom names) that need to be deleted in applying the patch

    See Also
    --------
    :class:`ResidueTemplate`

    Notes
    -----
    This class basically just provides an additional list of atoms that need to
    be deleted when applying this patch -- something that does not apply to
    standard Residues
    """
    def __init__(self, name=''):
        super(PatchTemplate, self).__init__(name)
        self.add_bonds = []
        self.delete_atoms = []
        self.delete_impropers = []

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class ResidueTemplateContainer(list):
    """
    A container of ResidueTemplate objects representing a unit with multiple
    residues

    Parameters
    ----------
    name : str, optional
        The name of the residue container
    """
    def __init__(self, name=''):
        self.box = None
        self.name = name

    @classmethod
    def from_structure(cls, struct, term_decorate=True):
        """
        Instantiates a ResidueTemplateContainer from a Structure instance filled
        with residues

        Parameters
        ----------
        struct : :class:`parmed.structure.Structure`
            The structure from which to generate the ResidueTemplateContainer
            from
        term_decorate : bool, optional
            If True, terminal amino and nucleic acid residues will be adorned as
            follows:

                * N-prepended if it is an N-terminal amino acid
                * C-prepended if it is a C-terminal amino acid
                * 5-appended if it is a 5'-terminal nucleic acid
                * 3-appended if it is a 3'-terminal nucleic acid

            For example, an N-terminal GLY will become NGLY, while a 5'-terminal
            DA will become DA5. Default is True
        """
        inst = cls()
        for res in struct.residues:
            rt = ResidueTemplate.from_residue(res)
            # See if we need to decorate the termini names
            if rt.head is None and rt.tail is not None and term_decorate:
                if AminoAcidResidue.has(rt.name):
                    if len(rt.name) != 4 or rt.name[0] != 'N':
                        rt.name = 'N%s' % rt.name
                elif RNAResidue.has(rt.name) or DNAResidue.has(rt.name):
                    if rt.name[-1] != '5':
                        rt.name = '%s5' % rt.name
            elif rt.tail is None and rt.head is not None and term_decorate:
                if AminoAcidResidue.has(rt.name):
                    if len(rt.name) != 4 or rt.name[0] != 'C':
                        rt.name = 'C%s' % rt.name
                elif RNAResidue.has(rt.name) or DNAResidue.has(rt.name):
                    if rt.name[-1] != '3':
                        rt.name = '%s3' % rt.name
            inst.append(rt)
        inst.box = struct.box
        return inst

    def __getitem__(self, value):
        if isinstance(value, str):
            # Behave like a dict here... albeit a slow one
            for res in self:
                if res.name == value: return res
        return list.__getitem__(self, value)

    def fix_charges(self, precision=4):
        """
        Adjusts the net charge of all residues in this ResidueContainer to match
        the closest integer charge

        Parameters
        ----------
        precision : int, optional
            The number of decimal places that each charge should be rounded to.
            Default is 4

        Returns
        -------
        self : :class:`ResidueTemplateContainer`
            The current residue template container whose ResidueTemplates are
            being modified

        Notes
        -----
        This method modifies everything in-place.

        Raises
        ------
        ValueError
            If you try to call fix_charges on a container with no templates
        """
        if len(self) == 0:
            raise ValueError('Cannot fix charges on an empty container')
        for res in self:
            res.fix_charges(precision=precision)
        return self

    def to_library(self):
        """
        Converts the ResidueTemplateContainer instance to a library of unique
        :class:`ResidueTemplate` instances. The first of each kind of residue is
        taken

        Returns
        -------
        residues : dict {str : :class:`ResidueTemplate`}
            The residue library with all residues from this residue collection
        """
        ret = OrderedDict()
        for res in self:
            if res.name in ret: continue
            ret[res.name] = res
        return ret

    @classmethod
    def from_library(cls, library, copy=False):
        """
        Converts a dictionary of ResidueTemplate items into a
        ResidueTemplateContainer.

        Parameters
        ----------
        library : dict or OrderedDict
            The library of ResidueTemplate objects to add to this container
        copy : bool, optional
            If True, copies of each ResidueTemplate in library is added to the
            ResidueTemplateContainer. Default is False

        Returns
        -------
        cont : ResidueTemplateContainer
            A ResidueTemplateContainer containing all of the residues defined in
            ``library``

        Notes
        -----
        If the library is ordered, that order is maintained

        Raises
        ------
        TypeError if any of the items in the input library is not a
        ResidueTemplate instance (or an instance of a subclass)
        """
        cont = cls()
        for _, res in iteritems(library):
            if not isinstance(res, ResidueTemplate):
                raise ValueError('%r is not a ResidueTemplate instance' % res)
            if copy:
                cont.append(_copy.copy(res))
            else:
                cont.append(res)
        return cont

    def save(self, fname, format=None, **kwargs):
        """
        Saves the current ResidueTemplateContainer in the requested file format.
        Supported formats can be specified explicitly or determined by file-name
        extension. The following formats are supported, with the recognized
        suffix and ``format`` keyword shown in parentheses:

            - MOL2 (.mol2)
            - MOL3 (.mol3)
            - OFF (.lib/.off)

        Parameters
        ----------
        fname : str
            Name of the file to save. If ``format`` is ``None`` (see below), the
            file type will be determined based on the filename extension. If the
            type cannot be determined, a ValueError is raised.
        format : str, optional
            The case-insensitive keyword specifying what type of file ``fname``
            should be saved as. If ``None`` (default), the file type will be
            determined from filename extension of ``fname``
        kwargs : keyword-arguments
            Remaining arguments are passed on to the file writing routines that
            are called by this function

        Raises
        ------
        ValueError if either filename extension or ``format`` are not recognized
        TypeError if the structure cannot be converted to the desired format for
        whatever reason

        Notes
        -----
        Mol2 and Mol3 files are saved as concatenated multiple @<MOLECULE>s. By
        contrast, ``Structure.save`` will save a single @<MOLECULE> mol2 file
        with multiple residues if the mol2 format is requested.
        """
        from parmed.amber.offlib import AmberOFFLibrary
        from parmed.formats.mol2 import Mol2File
        extmap = {
                '.mol2' : 'MOL2',
                '.mol3' : 'MOL3',
                '.off' : 'OFFLIB',
                '.lib' : 'OFFLIB',
        }
        if format is not None:
            format = format.upper()
        else:
            base, ext = os.path.splitext(fname)
            if ext in ('.bz2', '.gz'):
                ext = os.path.splitext(base)[1]
            if ext in extmap:
                format = extmap[ext]
            else:
                raise ValueError('Could not determine file type of %s' % fname)
        if format == 'MOL2':
            Mol2File.write(self, fname, mol3=False, split=True, **kwargs)
        elif format == 'MOL3':
            Mol2File.write(self, fname, mol3=True, split=True, **kwargs)
        elif format in ('OFFLIB', 'OFF'):
            AmberOFFLibrary.write(self.to_library(), fname, **kwargs)
        else:
            raise ValueError('Unrecognized format for ResidueTemplate save')



"""
This module contains the functionality for carrying out geometrical calculations
for molecules and molecular systems

Author: Jason Swails
Contributors:

Copyright (C) 2014 - 2015  Jason Swails

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your option) any
later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
59 Temple Place - Suite 330
Boston, MA 02111-1307, USA.
"""
from __future__ import division

from collections import defaultdict
from parmed import unit as u
from parmed.constants import TINY, DEG_TO_RAD, RAD_TO_DEG
from parmed.vec3 import Vec3
from math import pi, cos, sin, sqrt, acos
import numpy as np
import warnings

def box_lengths_and_angles_to_vectors(a, b, c, alpha, beta, gamma):
    """
    This function takes the lengths of the unit cell vectors and the angles
    between them and returns 3 unit cell vectors satisfying those dimensions

    Parameters
    ----------
    a : double (or length Quantity)
        Length of the first unit cell vector
    b : double (or length Quantity)
        Length of the second unit cell vector
    c : double (or length Quantity)
        Length of the third unit cell vector
    alpha : double (or angle Quantity)
        Angle between vectors b and c
    beta : double (or angle Quantity)
        Angle between vectors a and c
    gamma : double (or angle Quantity)
        Angle between vectors a and b

    Returns
    -------
    list Quantity, list Quantity, list Quantity
        The 3, 3-element vectors as quantities with dimension length

    Notes
    -----
    The unit cell lengths are assumed to be Angstroms if no explicit unit is
    given. The angles are assumed to be degrees
    """
    if u.is_quantity(a): a = a.value_in_unit(u.angstroms)
    if u.is_quantity(b): b = b.value_in_unit(u.angstroms)
    if u.is_quantity(c): c = c.value_in_unit(u.angstroms)
    if u.is_quantity(alpha): alpha = alpha.value_in_unit(u.degrees)
    if u.is_quantity(beta): beta = beta.value_in_unit(u.degrees)
    if u.is_quantity(gamma): gamma = gamma.value_in_unit(u.degrees)

    if alpha <= 2*pi and beta <= 2*pi and gamma <= 2*pi:
        warnings.warn('Strange unit cell vector angles detected. They '
                      'appear to be in radians...')

    alpha *= DEG_TO_RAD
    beta *= DEG_TO_RAD
    gamma *= DEG_TO_RAD

    av = [a, 0.0, 0.0]
    bx = b * cos(gamma)
    by = b * sin(gamma)
    bv = [bx, by, 0.0]
    cx = c * cos(beta)
    cy = c * (cos(alpha) - cos(beta)*cos(gamma)) / sin(gamma)
    cz = sqrt(c*c - cx*cx - cy*cy)
    cv = [cx, cy, cz]

    # Make sure that any tiny components are exactly 0
    if abs(bx) < TINY: bv[0] = 0
    if abs(by) < TINY: bv[1] = 0
    if abs(cx) < TINY: cv[0] = 0
    if abs(cy) < TINY: cv[1] = 0
    if abs(cz) < TINY: cv[2] = 0

    return (av, bv, cv) * u.angstroms

def box_vectors_to_lengths_and_angles(a, b, c):
    """
    This function takes the lengths of the unit cell vectors and the angles
    between them and returns 3 unit cell vectors satisfying those dimensions

    Parameters
    ----------
    a : collection of 3 floats (or length Quantity)
        The first unit cell vector
    b : collection of 3 floats (or length Quantity)
        The second unit cell vector
    c : collection of 3 floats (or length Quantity)
        The third unit cell vector

    Returns
    -------
    (a, b, c), (alpha, beta, gamma)
        Two tuples, the first is the 3 unit cell vector lengths as
        length-dimension Quantity objects and the second is the set of angles
        between the unit cell vectors as angle-dimension Quantity objects

    Notes
    -----
    The unit cell lengths are assumed to be Angstroms if no explicit unit is
    given.
    """
    if u.is_quantity(a): a = a.value_in_unit(u.angstroms)
    if u.is_quantity(b): b = b.value_in_unit(u.angstroms)
    if u.is_quantity(c): c = c.value_in_unit(u.angstroms)
    # Get the lengths
    la = sqrt(a[0]*a[0] + a[1]*a[1] + a[2]*a[2])
    lb = sqrt(b[0]*b[0] + b[1]*b[1] + b[2]*b[2])
    lc = sqrt(c[0]*c[0] + c[1]*c[1] + c[2]*c[2])
    # Angles
    alpha = acos((b[0]*c[0] + b[1]*c[1] + b[2]*c[2]) / (lb*lc))
    beta = acos((a[0]*c[0] + a[1]*c[1] + a[2]*c[2]) / (la*lc))
    gamma = acos((b[0]*a[0] + b[1]*a[1] + b[2]*a[2]) / (lb*la))
    # Convert to degrees
    alpha *= RAD_TO_DEG
    beta *= RAD_TO_DEG
    gamma *= RAD_TO_DEG

    return (la, lb, lc) * u.angstroms, (alpha, beta, gamma) * u.degrees

def reduce_box_vectors(a, b, c):
    """
    This function puts three unit cell vectors in a reduced form where a is
    "mostly" in x, b is "mostly" in y, and c is "mostly" in z. This form is
    necessary for some programs (notably OpenMM and Gromacs)

    Parameters
    ----------
    a : 3-element collection of float
        First unit cell vector
    b : 3-element collection of float
        Second unit cell vector
    c : 3-element collection of float
        Third unit cell vector

    Returns
    -------
    red_a, red_b, red_c : Vec3, Vec3, Vec3
        The reduced unit cell vectors in units of angstroms

    Notes
    -----
    The implementation here is taken from the OpenMM Python application layer
    written by Peter Eastman
    """
    if u.is_quantity(a):
        a = a.value_in_unit(u.angstroms)
    if u.is_quantity(b):
        b = b.value_in_unit(u.angstroms)
    if u.is_quantity(c):
        c = c.value_in_unit(u.angstroms)

    a = Vec3(*a)
    b = Vec3(*b)
    c = Vec3(*c)

    c = c - b*round(c[1]/b[1])
    c = c - a*round(c[0]/a[0])
    b = b - a*round(b[0]/a[0])

    return a, b, c


def center_of_mass(coordinates, masses):
    """ Compute the center of mass of a group of coordinates.

    Parameters
    ----------
    coordinates : numpy.ndarray
        Coordinate array
    masses : numpy.ndarray
        Array of masses

    Returns
    -------
    COM
        np.ndarray of shape (3,) identifying the cartesian center of mass

    Notes
    -----
    This method *requires* that the parameters be passed in as numpy arrays.
    AttributeError's will ensue if this is not the case. Also, coordinates must
    be able to be reshaped to (len(masses), 3), or ValueError's will ensue
    """
    masses = masses.flatten()
    coordinates = coordinates.reshape((masses.shape[0], 3))
    return np.average(coordinates, weights=masses, axis=0)

def distance2(a1, a2):
    """ Computes the cartesian distance between two atoms. Ignores periodic boundary conditions.

    Parameters
    ----------
    a1, a2 : Atom or collection of 3 coordinates
        The two atoms between whom the distance should be calculated

    Returns
    -------
    d2 : float
        The square of the distance between the two atoms

    Notes
    -----
    This is done in pure Python, so it should not be used for large numbers of
    distance calculations. For that, use numpy-vectorized routines and the numpy
    coordinate arrays

    Raises
    ------
    TypeError if a1 or a2 are not Atom or iterable
    ValueError if a1 or a2 are iterable, but do not have exactly 3 items
    """
    x1, y1, z1 = _get_coords_from_atom_or_tuple(a1)
    x2, y2, z2 = _get_coords_from_atom_or_tuple(a2)
    dx = x1 - x2
    dy = y1 - y2
    dz = z1 - z2
    return dx*dx + dy*dy + dz*dz

def angle(a1, a2, a3):
    """ Computes the cartesian angle between three atoms. Ignores periodic boundary conditions.

    Parameters
    ----------
    a1, a2, a3 : Atom or collection of 3 coordinates
        The three atoms between whom the angle should be calculated (with a2
        being the central atoms)

    Returns
    -------
    ang : float
        The angle between the vectors a1-a2 and a2-a3 in degrees

    Notes
    -----
    This is done in pure Python, so it should not be used for large numbers of
    distance calculations. For that, use numpy-vectorized routines and the numpy
    coordinate arrays

    Raises
    ------
    TypeError if a1, a2, or a3 are not Atom or iterable
    ValueError if a1, a2, or a3 are iterable, but do not have exactly 3 items
    """
    x1, y1, z1 = _get_coords_from_atom_or_tuple(a1)
    x2, y2, z2 = _get_coords_from_atom_or_tuple(a2)
    x3, y3, z3 = _get_coords_from_atom_or_tuple(a3)
    v1 = np.array([x2 - x1, y2 - y1, z2 - z1])
    v2 = np.array([x2 - x3, y2 - y3, z2 - z3])
    l1 = np.sqrt(np.dot(v1, v1))
    l2 = np.sqrt(np.dot(v2, v2))
    cosa = np.dot(v1, v2) / (l1 * l2)
    return np.degrees(np.arccos(cosa))

def dihedral(a1, a2, a3, a4):
    """
    Computes the angle between three vectors made up of four points (all three
    vectors share one point with one other vector)

    Parameters
    ----------
    a1, a2, a3, a4 : Atom or collection of 4 coordinates
        The four atoms between whom the torsion angle should be calculated (with
        a1 and a4 being the two end-point atoms not shared between two vectors)

    Returns
    -------
    dihed : float
        The measured dihedral between the 4 points in degrees
    """
    p = np.array([_get_coords_from_atom_or_tuple(a1),
                  _get_coords_from_atom_or_tuple(a2),
                  _get_coords_from_atom_or_tuple(a3),
                  _get_coords_from_atom_or_tuple(a4)])
    v1 = p[1] - p[0]
    v2 = p[1] - p[2]
    v3 = p[3] - p[2]
    # Take the cross product between v1-v2 and v2-v3
    v1xv2 = _cross(v1, v2)
    v2xv3 = _cross(v2, v3)
    # Now find the angle between these cross-products
    l1 = np.sqrt(np.dot(v1xv2, v1xv2))
    l2 = np.sqrt(np.dot(v2xv3, v2xv3))
    cosa = np.dot(v1xv2, v2xv3) / (l1 * l2)
    if np.dot(v3, v1xv2) <= 0.0 :
        return np.degrees(np.arccos(cosa))
    else :
        return -np.degrees(np.arccos(cosa))

def _cross(v1, v2):
    """ Computes the cross-product """
    # Can't use np.cross for pypy, since it's not yet implemented
    return np.array([v1[1]*v2[2] - v1[2]*v2[1],
                     v1[2]*v2[0] - v1[0]*v2[2],
                     v1[0]*v2[1] - v1[1]*v2[0]])


def _get_coords_from_atom_or_tuple(a):
    from parmed.topologyobjects import Atom
    if isinstance(a, Atom):
        return a.xx, a.xy, a.xz
    return a

# tuples are pairs of atomic numbers followed by the distance cutoff below which
# they are considered "bonded". This is taken from Atom.cpp in cpptraj
# (Atom::GetBondLength)
_OFFSET = 0.20 # offset for what is considered a bond
_DEFAULT_CUTOFF = (1.6 + _OFFSET)**2
STANDARD_BOND_LENGTHS_SQUARED = defaultdict(lambda: _DEFAULT_CUTOFF)
STANDARD_BOND_LENGTHS_SQUARED.update({
    # Self-bonds
    (1, 1) : (0.74 + _OFFSET)**2,
    (6, 6) : (1.54 + _OFFSET)**2,
    (7, 7) : (1.45 + _OFFSET)**2,
    (8, 8) : (1.48 + _OFFSET)**2,
    (15, 15) : (2.21 + _OFFSET)**2,
    (16, 16) : (2.05 + _OFFSET)**2,
    # H-
    (1, 6) : (1.09 + _OFFSET)**2, (6, 1) : (1.09 + _OFFSET)**2,
    (1, 7) : (1.01 + _OFFSET)**2, (7, 1) : (1.01 + _OFFSET)**2,
    (1, 8) : (0.96 + _OFFSET)**2, (8, 1) : (0.96 + _OFFSET)**2,
    (1, 15) : (1.44 + _OFFSET)**2, (15, 1) : (1.44 + _OFFSET)**2,
    (1, 16) : (1.34 + _OFFSET)**2, (16, 1) : (1.34 + _OFFSET)**2,
    # C-
    (6, 7) : (1.47 + _OFFSET)**2, (7, 6) : (1.47 + _OFFSET)**2,
    (6, 8) : (1.43 + _OFFSET)**2, (8, 6) : (1.43 + _OFFSET)**2,
    (6, 9) : (1.35 + _OFFSET)**2, (9, 6) : (1.35 + _OFFSET)**2,
    (6, 15) : (1.84 + _OFFSET)**2, (15, 6) : (1.84 + _OFFSET)**2,
    (6, 16) : (1.82 + _OFFSET)**2, (16, 6) : (1.82 + _OFFSET)**2,
    (6, 17) : (1.77 + _OFFSET)**2, (17, 6) : (1.77 + _OFFSET)**2,
    (6, 35) : (1.94 + _OFFSET)**2, (35, 6) : (1.94 + _OFFSET)**2,
    # N-
    (7, 8) : (1.40 + _OFFSET)**2, (8, 7) : (1.40 + _OFFSET)**2,
    (7, 9) : (1.36 + _OFFSET)**2, (9, 7) : (1.36 + _OFFSET)**2,
    (7, 15) : (1.71 + _OFFSET)**2, (15, 7) : (1.71 + _OFFSET)**2,
    (7, 16) : (1.68 + _OFFSET)**2, (16, 7) : (1.68 + _OFFSET)**2,
    (7, 17) : (1.75 + _OFFSET)**2, (17, 7) : (1.75 + _OFFSET)**2,
    # O-
    (8, 9) : (1.42 + _OFFSET)**2, (9, 8) : (1.42 + _OFFSET)**2,
    (8, 15) : (1.63 + _OFFSET)**2, (15, 8) : (1.63 + _OFFSET)**2,
    (8, 16) : (1.48 + _OFFSET)**2, (16, 8) : (1.48 + _OFFSET)**2,
    # F-
    (9, 15) : (1.54 + _OFFSET)**2, (15, 9) : (1.54 + _OFFSET)**2,
    (9, 16) : (1.56 + _OFFSET)**2, (16, 9) : (1.56 + _OFFSET)**2,
    # P-
    (15, 16) : (1.86 + _OFFSET)**2, (16, 15) : (1.86 + _OFFSET)**2,
    (15, 17) : (2.03 + _OFFSET)**2, (17, 15) : (2.03 + _OFFSET)**2,
    # S-
    (16, 17) : (2.07 + _OFFSET)**2, (17, 16) : (2.07 + _OFFSET)**2,
})



"""
This package contains classes responsible for loading rdkit objects
"""

from __future__ import print_function, absolute_import
from parmed.formats import PDBFile
from parmed.utils.six.moves import StringIO


class RDKit(object):

    @staticmethod
    def load(rmol):
        """
        Load a :class:`Mol` object and return a populated :class:`Structure`
        instance

        Parameters
        ----------
        rmol: :class:`Mol`
            RDKit :class:`Mol` object to convert

        Examples
        --------
        >>> from rdkit import Chem
        >>> import parmed as pmd
        >>> mol = Chem.MolFromSmiles('Cc1ccccc1')
        >>> struct = pmd.load_rdkit(mol)
        """
        from rdkit import Chem
        fh = StringIO(Chem.MolToPDBBlock(rmol))
        return PDBFile.parse(fh)

    @staticmethod
    def from_smiles(smiles, coordinates=True):
        """
        Load smiles string to :class:`Structure`

        Parameters
        ----------
        smiles : str, smiles
        coordinates : bool, default True
            if True, use `rdkit.Chem.AllChem.EmbedMultipleConfs to assign coordinates

        Returns
        -------
        parm : :class:`Structure`
        """
        from rdkit import Chem
        from rdkit.Chem import AllChem
        mol = Chem.MolFromSmiles(smiles)

        if coordinates:
            AllChem.EmbedMultipleConfs(mol, useExpTorsionAnglePrefs=True,
                    useBasicKnowledge=True)

        parm = RDKit.load(mol)
        if not coordinates:
            parm.coordinates = None
            parm._coordinates = None
        return parm

    @staticmethod
    def from_sdf(filename, structure=False):
        """
        Load SDF file to :class:`Structure`

        Parameters
        ----------
        filename: str
        structure : bool, default False
            if True, return a :class:`Structure`
            if False, return a list of :class:`Structure`
        """
        from rdkit import Chem
        sdf_collection = Chem.SDMolSupplier(filename, removeHs=False)
        if structure:
            mol = next(sdf_collection)
            return RDKit.load(mol)
        else:
            return [RDKit.load(mol) for mol in sdf_collection]



"""
This module contains classes for parsing and processing CHARMM parameter,
topology, and stream files.

Author: Jason M. Swails
Contributors:
Date: Mar. 26, 2017
"""
from __future__ import absolute_import, division, print_function

import os
import re
import warnings
from copy import copy as _copy
from collections import OrderedDict
from itertools import combinations

from ..constants import TINY
from ..exceptions import CharmmError, ParameterWarning
from ..modeller import PatchTemplate, ResidueTemplate
from ..parameters import ParameterSet
from ..periodic_table import AtomicNum, element_by_mass
from ..topologyobjects import (AngleType, Atom, AtomType, BondType, CmapType,
                               DihedralType, DihedralTypeList, ImproperType,
                               NoUreyBradley, DrudeAtom, DrudeAnisotropy)
from ..utils.io import genopen
from ..utils.six import integer_types, iteritems, string_types
from ..utils.six.moves import zip
from ._charmmfile import CharmmFile, CharmmStreamFile

_penaltyre = re.compile(r'penalty\s*=\s*([\d\.]+)')

class _EmptyStringIterator(object):
    """ Always yields an empty string """
    def __iter__(self):
        while True:
            yield ''
    def __getitem__(self, idx):
        return ''

def _typeconv(name):
    if isinstance(name, integer_types):
        return name
    if name.upper() == name:
        return name.replace('*', 'STR').replace('+', 'P').replace('-', 'M')[:6]
    # Lowercase letters present -- decorate the type name with LTU --
    # Lower To Upper
    return ('%sLTU' % name.upper()).replace('*', 'STR').replace(
            '+', 'P').replace('-', 'M')[:6]

class CharmmImproperMatchingMixin(object):
    """ Implements CHARMM-style improper matching """

    def match_improper_type(self, a1, a2, a3, a4):
        """ Matches an improper type based on atom type names """
        typ = self._match_improper_with_typemap(self.improper_types, a1, a2, a3, a4)
        if typ is None:
            typ = self._match_improper_with_typemap(self.improper_periodic_types, a1, a2, a3, a4)
        return typ

    def _match_improper_with_typemap(self, typemap, a1, a2, a3, a4):

        if (a1, a2, a3, a4) in typemap: return typemap[(a1, a2, a3, a4)]
        if (a4, a3, a2, a1) in typemap: return typemap[(a4, a3, a2, a1)]

        # Now try any of the sortings. The documented CHARMM ordering does not seem to work for
        # all systems CHARMM supports :(

        key = tuple(sorted([a1, a2, a3, a4]))
        if self._improper_key_map.get(key, None) in typemap:
            return typemap[self._improper_key_map[key]]

        for exact1, exact2, exact3 in combinations((a1, a2, a3, a4), 3):
            key = tuple(sorted([exact1, exact2, exact3, 'X']))
            if self._improper_key_map.get(key, None) in typemap:
                return typemap[self._improper_key_map[key]]

        for exact1, exact2 in combinations((a1, a2, a3, a4), 2):
            key = tuple(sorted([exact1, exact2, 'X', 'X']))
            if self._improper_key_map.get(key, None) in typemap:
                return typemap[self._improper_key_map[key]]

        return None

class CharmmParameterSet(ParameterSet, CharmmImproperMatchingMixin):
    """
    Stores a parameter set defined by CHARMM files. It stores the equivalent of
    the information found in the MASS section of the CHARMM topology file
    (TOP/RTF) and all of the information in the parameter files (PAR)

    Parameters
    ----------
    *filenames : variable length arguments of str
        The list of topology, parameter, and stream files to load into the
        parameter set. The following file type suffixes are recognized:
            .rtf, .top -- Residue topology file
            .par, .prm -- Parameter file
            .str -- Stream file
            .inp -- If "par" is in the file name, it is a parameter file. If
                    "top" is in the file name, it is a topology file.
                    Otherwise, ValueError is raised.

    See Also
    --------
    :class:`parmed.parameters.ParameterSet`
    """

    def __copy__(self):
        other = super(CharmmParameterSet, self).__copy__()
        other._declared_nbrules = self._declared_nbrules
        return other

    @staticmethod
    def _convert(data, type, msg='', line_index=None, line=None):
        """
        Converts a data type to a desired type, raising CharmmError if it
        fails
        """
        try:
            return type(data)
        except ValueError:
            msg = 'Could not convert %s to %s\n' % (msg, type)
            if line_index is not None:
                msg += 'input line %d\n' % line_index
            if line is not None:
                msg += 'input line: %s\n' % line
            raise CharmmError(msg)

    def __init__(self, *args):
        # Instantiate the list types
        super(CharmmParameterSet, self).__init__()
        self.parametersets = []
        self._declared_nbrules = False

        # Load all of the files
        tops, pars, strs = [], [], []
        for arg in args:
            if arg.endswith('.rtf') or arg.endswith('.top'):
                tops.append(arg)
            elif arg.endswith('.par') or arg.endswith('.prm'):
                pars.append(arg)
            elif arg.endswith('.str'):
                strs.append(arg)
            elif arg.endswith('.inp'):
                # Only consider the file name (since the directory is likely
                # "toppar" and will screw up file type detection)
                fname = os.path.split(arg)[1]
                if 'par' in fname:
                    pars.append(arg)
                elif 'top' in fname:
                    tops.append(arg)
                else:
                    raise ValueError('Unrecognized file type: %s' % arg)
            else:
                raise ValueError('Unrecognized file type: %s' % arg)
        for top in tops:
            self.read_topology_file(top)
        for par in pars:
            self.read_parameter_file(par)
        for strf in strs:
            self.read_stream_file(strf)

    @classmethod
    def from_parameterset(cls, params, copy=False):
        """
        Instantiates a CharmmParameterSet from another ParameterSet (or
        subclass). The main thing this feature is responsible for is converting
        lower-case atom type names into all upper-case and decorating the name
        to ensure each atom type name is unique.

        Parameters
        ----------
        params : :class:`parmed.parameters.ParameterSet`
            ParameterSet containing the list of parameters to be converted to a
            CHARMM-compatible set
        copy : bool, optional
            If True, the returned parameter set is a deep copy of ``params``. If
            False, the returned parameter set is a shallow copy, and the
            original set may be modified if any lower-case atom type names are
            present. Default is False.

        Returns
        -------
        new_params : CharmmParameterSet
            The parameter set whose atom type names are converted to all
            upper-case
        """
        new_params = cls()
        if copy:
            do_copy = lambda x: _copy(x)
        else:
            do_copy = lambda x: x
        # Convert all parameters
        id_typemap = dict()
        def copy_paramtype(key, typ, dict):
            if isinstance(key, string_types):
                key = _typeconv(key)
            elif isinstance(key, tuple):
                key = tuple(_typeconv(k) for k in key)
            # NoUreyBradley should never be copied
            if typ is NoUreyBradley:
                dict[key] = NoUreyBradley
            elif id(typ) in id_typemap:
                dict[key] = id_typemap[id(typ)]
            else:
                newtype = do_copy(typ)
                id_typemap[id(typ)] = newtype
                dict[key] = newtype

        for key, atom_type in iteritems(params.atom_types_tuple):
            atom_type.name = _typeconv(atom_type.name)
            copy_paramtype(key, atom_type, new_params.atom_types_tuple)
        for typename, atom_type in iteritems(params.atom_types):
            atom_type.name = _typeconv(atom_type.name)
            copy_paramtype(typename, atom_type, new_params.atom_types)
        for idx, atom_type in iteritems(params.atom_types_int):
            atom_type.name = _typeconv(atom_type.name)
            copy_paramtype(idx, atom_type, new_params.atom_types_int)

        for key, typ in iteritems(params.bond_types):
            copy_paramtype(key, typ, new_params.bond_types)
        for key, typ in iteritems(params.angle_types):
            copy_paramtype(key, typ, new_params.angle_types)
        for key, typ in iteritems(params.urey_bradley_types):
            copy_paramtype(key, typ, new_params.urey_bradley_types)
        for key, typ in iteritems(params.dihedral_types):
            copy_paramtype(key, typ, new_params.dihedral_types)
        for key, typ in iteritems(params.improper_periodic_types):
            copy_paramtype(key, typ, new_params.improper_periodic_types)
        for key, typ in iteritems(params.improper_types):
            copy_paramtype(key, typ, new_params.improper_types)
        for key, typ in iteritems(params.cmap_types):
            assert len(key) == 8, '%d-key cmap type detected!' % len(key)
            copy_paramtype(key, typ, new_params.cmap_types)
        for key, typ in iteritems(params.nbfix_types):
            copy_paramtype(key, typ, new_params.nbfix_types)

        return new_params

    @classmethod
    def from_structure(cls, struct):
        """ Extracts known parameters from a Structure instance

        Parameters
        ----------
        struct : :class:`parmed.structure.Structure`
            The parametrized ``Structure`` instance from which to extract
            parameters into a ParameterSet

        Returns
        -------
        params : :class:`ParameterSet`
            The parameter set with all parameters defined in the Structure

        Notes
        -----
        The parameters here are copies of the ones in the Structure, so
        modifying the generated ParameterSet will have no effect on ``struct``.
        Furthermore, the *first* occurrence of each parameter will be used. If
        future ones differ, they will be silently ignored, since this is
        expected behavior in some instances (like with Gromacs topologies in the
        ff99sb-ildn force field).

        Dihedrals are a little trickier. They can be multi-term, which can be
        represented either as a *single* entry in dihedrals with a type of
        DihedralTypeList or multiple entries in dihedrals with a DihedralType
        parameter type. In this case, the parameter is constructed from either
        the first DihedralTypeList found or the first DihedralType of each
        periodicity found if no matching DihedralTypeList is found.
        """
        return cls.from_parameterset(
            ParameterSet.from_structure(struct, allow_unequal_duplicates=False)
        )

    @classmethod
    def load_set(cls, tfile=None, pfile=None, sfiles=None):
        """
        Instantiates a CharmmParameterSet from a Topology file and a Parameter
        file (or just a Parameter file if it has all information)

        Parameters
        ----------
        tfile : str
            The name of the Topology (RTF/TOP) file to parse
        pfile : str
            The name of the Parameter (PAR/PRM) file to parse
        sfiles : list(str)
            Iterable of stream (STR) file names

        Returns
        -------
        New CharmmParameterSet populated with parameters found in the provided
        files

        Notes
        -----
        The RTF file is read first (if provided), followed by the PAR file,
        followed by the list of stream files (in the order they are provided).
        Parameters in each stream file will overwrite those that came before (or
        simply append to the existing set if they are different)
        """
        inst = cls()
        if tfile is not None:
            inst.read_topology_file(tfile)
        if pfile is not None:
            inst.read_parameter_file(pfile)
        if isinstance(sfiles, string_types):
            # The API docstring requests a list, but allow for users to pass a
            # string with a single filename instead
            inst.read_stream_file(sfiles)
        elif sfiles is not None:
            for sfile in sfiles:
                inst.read_stream_file(sfile)
        return inst

    def read_parameter_file(self, pfile, comments=None):
        """
        Reads all of the parameters from a parameter file. Versions 36 and
        later of the CHARMM force field files have an ATOMS section defining
        all of the atom types.  Older versions need to load this information
        from the RTF/TOP files.

        Parameters
        ----------
        pfile : str or list of lines
            Name of the CHARMM parameter file to read or list of lines to parse
            as a file
        comments : list of str, optional
            List of comments on each of the pfile lines (if pfile is a list of
            lines)

        Notes
        -----
        The atom types must all be loaded by the end of this routine.  Either
        supply a PAR file with atom definitions in them or read in a RTF/TOP
        file first. Failure to do so will result in a raised RuntimeError.
        """
        conv = CharmmParameterSet._convert
        if isinstance(pfile, str):
            own_handle = True
            f = CharmmFile(pfile)
        else:
            own_handle = False
            f = pfile
            if not isinstance(f, CharmmFile) and comments is None:
                comments = _EmptyStringIterator()
        # What section are we parsing?
        section = None
        # The current cmap we are building (these span multiple lines)
        current_cmap = None
        current_cmap2 = None
        current_cmap_data = []
        current_cmap_res = 0
        nonbonded_types = dict() # Holder
        parameterset = None
        declared_geometric = False
        for i, line in enumerate(f):
            line = line.strip()
            try:
                comment = f.comment
            except AttributeError:
                comment = comments[i]
            if not line:
                # This is a blank line
                continue
            if parameterset is None and line.strip().startswith('*>>'):
                parameterset = line.strip()[1:78]
                continue
            # Set section if this is a section header
            if line.upper().startswith('ATOM'):
                section = 'ATOMS'
                continue
            if line.upper().startswith('BOND'):
                section = 'BONDS'
                continue
            if line.upper().startswith('ANGLE') or line.upper().startswith('THETA'):
                section = 'ANGLES'
                continue
            if line.upper().startswith('DIHE') or line.upper().startswith('PHI'):
                section = 'DIHEDRALS'
                continue
            if line.upper().startswith('IMPROPER') or line.upper().startswith('IMPHI'):
                section = 'IMPROPER'
                continue
            if line.upper().startswith('CMAP'):
                section = 'CMAP'
                continue
            if line.upper().startswith('NONBONDED'):
                read_first_nonbonded = declared_geometric = False
                section = 'NONBONDED'
                # Get nonbonded keywords
                words = line.split()[1:]
                scee = None
                for i, word in enumerate(words):
                    if word.upper() == 'E14FAC':
                        try:
                            scee = 1 / float(words[i+1])
                        except (ValueError, IndexError):
                            raise CharmmError('Could not parse 1-4 electrostatic scaling factor '
                                              'from NONBONDED card')
                        if self._declared_nbrules:
                            if len(self.dihedral_types) > 0:
                                # We already specified it -- make sure it's the same
                                # as the one we specified before
                                _, dt0 = next(iteritems(self.dihedral_types))
                                diff = abs(dt0[0].scee - scee)
                                if diff > TINY:
                                    raise CharmmError('Inconsistent 1-4 scalings')
                        else:
                            for key, dtl in iteritems(self.dihedral_types):
                                for dt in dtl:
                                    dt.scee = scee
                    elif word.upper().startswith('GEOM'):
                        if self._declared_nbrules and self.combining_rule != 'geometric':
                            raise CharmmError('Cannot combine parameter files with different '
                                              'combining rules')
                        self.combining_rule = 'geometric'
                        declared_geometric = True
                continue
            if line.upper().startswith('NBFIX'):
                section = 'NBFIX'
                continue
            if line.upper().startswith('HBOND'):
                section = None
                continue
            if line.upper().startswith('THOLE'):
                section = None
                continue
            # It seems like files? sections? can be terminated with 'END'
            if line[:3].upper() == 'END':
                section = None
                continue
            # If we have no section, skip
            if section is None: continue
            # See if our comments define a penalty for this line
            pens = _penaltyre.findall(comment)
            if len(pens) == 1:
                penalty = float(pens[0])
            else:
                penalty = None
            # Now handle each section specifically
            if section.upper() == 'ATOMS':
                if not line.upper().startswith('MASS'): continue # Should this happen?
                words = line.split()
                if words[0].upper() == 'END':
                    continue
                try:
                    idx = conv(words[1], int, 'atom type', line_index=i, line=line)
                    name = words[2].upper()
                    mass = conv(words[3], float, 'atom mass', line_index=i, line=line)
                except IndexError:
                    raise CharmmError('Could not parse MASS section.')
                # The parameter file might or might not have an element name
                try:
                    elem = words[4].upper()
                    if len(elem) == 2:
                        elem = elem[0] + elem[1].lower()
                    atomic_number = AtomicNum[elem]
                except (IndexError, KeyError):
                    # Figure it out from the mass
                    atomic_number = AtomicNum[element_by_mass(mass)]
                atype = AtomType(name=name, number=idx, mass=mass, atomic_number=atomic_number)
                self.atom_types_str[atype.name] = atype
                self.atom_types_int[atype.number] = atype
                self.atom_types_tuple[(atype.name, atype.number)] = atype
                continue
            if section.upper() == 'BONDS':
                words = line.split()
                if words[0].upper() == 'END':
                    continue
                try:
                    type1 = words[0].upper()
                    type2 = words[1].upper()
                    k = conv(words[2], float, 'bond force constant', line_index=i, line=line)
                    req = conv(words[3], float, 'bond equilibrium dist', line_index=i, line=line)
                except IndexError:
                    raise CharmmError('Could not parse bonds.')
                key = (min(type1, type2), max(type1, type2))
                bond_type = BondType(k, req)
                if key in self.bond_types:
                    # See if existing bond type has a different value and replaces it with a warning
                    if self.bond_types[key] != bond_type:
                        # Replace. Warn if they are different
                        warnings.warn('Replacing bond %r, %r with %r' %
                                      (key, self.bond_types[key], bond_type), ParameterWarning)
                        self.bond_types[(type1, type2)] = bond_type
                        self.bond_types[(type2, type1)] = bond_type
                else: # key not present
                    self.bond_types[(type1, type2)] = bond_type
                    self.bond_types[(type2, type1)] = bond_type
                bond_type.penalty = penalty
                continue
            if section.upper() == 'ANGLES':
                words = line.split()
                if words[0].upper() == 'END':
                    continue
                try:
                    type1 = words[0].upper()
                    type2 = words[1].upper()
                    type3 = words[2].upper()
                    k = conv(words[3], float, 'angle force constant', line_index=i, line=line)
                    theteq = conv(words[4], float, 'angle equilibrium value', line_index=i, line=line)
                except IndexError:
                    raise CharmmError('Could not parse angles.')

                angle_type = AngleType(k, theteq)
                key = (type1, type2, type3)
                if key in self.angle_types:
                    # See if the existing angle type list has a different value
                    # and replaces it with a warning
                    if self.angle_types[key] != angle_type:
                        # Replace. Warn if they are different
                        warnings.warn('Replacing angle %r, %r with %r' %
                                      (key, self.angle_types[key], angle_type), ParameterWarning)
                        self.angle_types[(type1, type2, type3)] = angle_type
                        self.angle_types[(type3, type2, type1)] = angle_type
                else: # key not present
                    self.angle_types[(type1, type2, type3)] = angle_type
                    self.angle_types[(type3, type2, type1)] = angle_type
                # See if we have a urey-bradley
                try:
                    ubk = conv(words[5], float, 'Urey-Bradley force constant', line_index=i, line=line)
                    ubeq = conv(words[6], float, 'Urey-Bradley equil. value', line_index=i, line=line)
                    ubtype = BondType(ubk, ubeq)
                    ubtype.penalty = penalty
                except IndexError:
                    ubtype = NoUreyBradley
                self.urey_bradley_types[(type1, type2, type3)] = ubtype
                self.urey_bradley_types[(type3, type2, type1)] = ubtype
                angle_type.penalty = penalty
                continue
            if section.upper() == 'DIHEDRALS':
                words = line.split()
                if words[0].upper == 'END':
                    continue
                try:
                    type1 = words[0].upper()
                    type2 = words[1].upper()
                    type3 = words[2].upper()
                    type4 = words[3].upper()
                    k = conv(words[4], float, 'dihedral force constant', line_index=i, line=line)
                    n = conv(words[5], float, 'dihedral periodicity', line_index=i, line=line)
                    phase = conv(words[6], float, 'dihedral phase', line_index=i, line=line)
                except IndexError:
                    raise CharmmError('Could not parse dihedrals.')
                key = (type1, type2, type3, type4)
                # See if this is a second (or more) term of the dihedral group
                # that's already present.
                dihedral = DihedralType(k, n, phase)
                dihedral.penalty = penalty
                if key in self.dihedral_types:
                    # See if the existing dihedral type list has a term with
                    # the same periodicity -- If so, replace it
                    replaced = False
                    for i, dtype in enumerate(self.dihedral_types[key]):
                        if dtype.per == dihedral.per:
                            # Replace. Warn if they are different
                            if dtype != dihedral:
                                warnings.warn('Replacing dihedral %r with %r' % (dtype, dihedral),
                                              ParameterWarning)
                            self.dihedral_types[key][i] = dihedral
                            replaced = True
                            break
                    if not replaced:
                        self.dihedral_types[key].append(dihedral)
                else: # key not present
                    dtl = DihedralTypeList()
                    dtl.append(dihedral)
                    self.dihedral_types[(type1, type2, type3, type4)] = dtl
                    self.dihedral_types[(type4, type3, type2, type1)] = dtl
                continue
            if section.upper() == 'IMPROPER':
                words = line.split()
                if words[0].upper() == 'END':
                    continue
                try:
                    type1 = words[0].upper()
                    type2 = words[1].upper()
                    type3 = words[2].upper()
                    type4 = words[3].upper()
                    k = conv(words[4], float, 'improper force constant', line_index=i, line=line)
                    theteq = conv(words[5], float, 'improper equil. value', line_index=i, line=line)
                except IndexError:
                    raise CharmmError('Could not parse dihedrals.')
                # If we have a 7th column, that is the real psi0 (and the 6th
                # is the multiplicity, which will indicate this is a periodic
                # improper torsion (so it needs to be added to the
                # improper_periodic_types list)
                try:
                    tmp = conv(words[6], float, 'improper equil. value', line_index=i, line=line)
                except IndexError:
                    per = 0
                else:
                    per = int(theteq)
                    theteq = tmp
                # Improper types seem not to always have the central atom
                # defined in the first place, so just have the key a fully
                # sorted list. We still depend on the PSF having properly
                # ordered improper atoms
                key = (type1, type2, type3, type4)
                self._improper_key_map[tuple(sorted(key))] = key
                if per == 0:
                    improp = ImproperType(k, theteq)
                    self.improper_types[key] = improp
                else:
                    improp = DihedralType(k, per, theteq)
                    self.improper_periodic_types[key] = improp
                    improp.improper = True
                improp.penalty = penalty
                continue
            if section.upper() == 'CMAP':
                # This is the most complicated part, since cmap parameters span
                # many lines. We won't do much error catching here.
                words = line.split()
                if words[0].upper() == 'END':
                    continue
                try:
                    holder = [float(w) for w in words]
                    current_cmap_data.extend(holder)
                except ValueError:
                    # We assume this is a definition of a new CMAP, so
                    # terminate the last CMAP if applicable
                    if current_cmap is not None:
                        # We have a map to terminate
                        ty = CmapType(current_cmap_res, current_cmap_data)
                        self.cmap_types[current_cmap] = ty
                        self.cmap_types[current_cmap2] = ty
                    try:
                        type1 = words[0].upper()
                        type2 = words[1].upper()
                        type3 = words[2].upper()
                        type4 = words[3].upper()
                        type5 = words[4].upper()
                        type6 = words[5].upper()
                        type7 = words[6].upper()
                        type8 = words[7].upper()
                        res = conv(words[8], int, 'CMAP resolution', line_index=i, line=line)
                    except IndexError:
                        raise CharmmError('Could not parse CMAP data.')
                    # order the torsions independently
                    k1 = [type1, type2, type3, type4, type5, type6, type7, type8]
                    k2 = [type8, type7, type6, type5, type4, type3, type2, type1]
                    current_cmap = tuple(min(k1, k2))
                    current_cmap2 = tuple(max(k1, k2))
                    current_cmap_res = res
                    current_cmap_data = []
                continue
            if section.upper() == 'NONBONDED':
                # Now get the nonbonded values
                words = line.split()
                if words[0].upper == 'END':
                    continue
                try:
                    atype = words[0].upper()
                    # 1st column is ignored
                    epsilon = conv(words[2], float, 'vdW epsilon term', line_index=i, line=line)
                    rmin = conv(words[3], float, 'vdW Rmin/2 term', line_index=i, line=line)
                except (IndexError, CharmmError):
                    # If we haven't read our first nonbonded term yet, we may
                    # just be parsing the settings that should be used. So
                    # soldier on
                    if read_first_nonbonded: raise
                    for i, word in enumerate(words):
                        if word.upper() == 'E14FAC':
                            try:
                                scee = 1 / float(words[i+1])
                            except (ValueError, IndexError):
                                raise CharmmError('Could not parse electrostatic scaling constant')
                            if self._declared_nbrules:
                                # We already specified it -- make sure it's the
                                # same as the one we specified before
                                _, dt0 = next(iteritems(self.dihedral_types))
                                diff = abs(dt0[0].scee - scee)
                                if diff > TINY:
                                    raise CharmmError('Inconsistent 1-4 scalings')
                            else:
                                for key, dtl in iteritems(self.dihedral_types):
                                    for dt in dtl:
                                        dt.scee = scee
                        elif word.upper().startswith('GEOM'):
                            if self._declared_nbrules and self.combining_rule != 'geometric':
                                raise CharmmError('Cannot combine parameter files with different '
                                                  'combining rules')
                            self.combining_rule = 'geometric'
                            declared_geometric = True
                    continue
                else:
                    # OK, we've read our first nonbonded section for sure now.
                    # Make sure we did not try to read in a str file that did
                    # not define GEOM if a previous file did, since
                    # Lorentz-Berthelot and geometric combining rules are
                    # incompatible
                    if (self._declared_nbrules and self.combining_rule == 'geometric' and
                        not declared_geometric):
                        raise CharmmError('Cannot combine parameter files with '
                                          'different combining rules')
                    read_first_nonbonded = True
                    self._declared_nbrules = True
                # See if we have 1-4 parameters
                try:
                    # 4th column is ignored
                    eps14 = conv(words[5], float, '1-4 vdW epsilon term', line_index=i, line=line)
                    rmin14 = conv(words[6], float, '1-4 vdW Rmin/2 term', line_index=i, line=line)
                except IndexError:
                    eps14 = rmin14 = None
                nonbonded_types[atype] = [epsilon, rmin, eps14, rmin14]
                continue
            if section.upper() == 'NBFIX':
                words = line.split()
                if words[0].upper() == 'END':
                    continue
                try:
                    at1 = words[0].upper()
                    at2 = words[1].upper()
                    emin = abs(conv(words[2], float, 'NBFIX Emin', line_index=i, line=line))
                    rmin = conv(words[3], float, 'NBFIX Rmin', line_index=i, line=line)
                    try:
                        emin14 = abs(conv(words[4], float, 'NBFIX Emin 1-4', line_index=i, line=line))
                        rmin14 = conv(words[5], float, 'NBFIX Rmin 1-4', line_index=i, line=line)
                    except IndexError:
                        emin14 = rmin14 = None
                    try:
                        self.atom_types_str[at1].add_nbfix(at2, rmin, emin, rmin14, emin14)
                        self.atom_types_str[at2].add_nbfix(at1, rmin, emin, rmin14, emin14)
                    except KeyError:
                        # Some stream files define NBFIX terms with an atom that
                        # is defined in another toppar file that does not
                        # necessarily have to be loaded. As a result, not every
                        # NBFIX found here will necessarily need to be applied.
                        # If we can't find a particular atom type, don't bother
                        # adding that nbfix and press on
                        pass
                except IndexError:
                    raise CharmmError('Could not parse NBFIX terms.')
                self.nbfix_types[(min(at1, at2), max(at1, at2))] = (emin, rmin)
        # If we had any CMAP terms, then the last one will not have been added
        # yet. Add it here
        if current_cmap is not None:
            typ = CmapType(current_cmap_res, current_cmap_data)
            self.cmap_types[current_cmap] = typ
            self.cmap_types[current_cmap2] = typ
        # Now we're done. Load the nonbonded types into the relevant AtomType
        # instances. In order for this to work, all keys in nonbonded_types
        # must be in the self.atom_types_str dict. Raise a RuntimeError if this
        # is not satisfied
        try:
            for key in nonbonded_types:
                self.atom_types_str[key].set_lj_params(*nonbonded_types[key])
        except KeyError:
            warnings.warn('Atom type %s not present in AtomType list' % key, ParameterWarning)
        if parameterset is not None:
            self.parametersets.append(parameterset)
        if own_handle:
            f.close()

    def read_topology_file(self, tfile):
        """
        Reads _only_ the atom type definitions from a topology file. This is
        unnecessary for versions 36 and later of the CHARMM force field.

        Parameters
        ----------
        tfile : str
            Name of the CHARMM topology file to read
        """
        conv = CharmmParameterSet._convert
        if isinstance(tfile, str):
            own_handle = True
            f = iter(CharmmFile(tfile))
        else:
            own_handle = False
            f = tfile
        hpatch = tpatch = None # default Head and Tail patches
        residues = OrderedDict()
        patches = OrderedDict()
        hpatches = OrderedDict()
        tpatches = OrderedDict()
        line = next(f)
        line_index = 0
        skip_adding_residue = False
        try:
            while line:
                line = line.strip()
                if line[:4].upper() == 'MASS':
                    words = line.split()
                    try:
                        idx = conv(words[1], int, 'atom type', line_index=line_index, line=line)
                        name = words[2].upper()
                        mass = conv(words[3], float, 'atom mass', line_index=line_index, line=line)
                    except IndexError:
                        raise CharmmError('Could not parse MASS section of %s' % tfile)
                    # The parameter file might or might not have an element name
                    try:
                        elem = words[4].upper()
                        if len(elem) == 2:
                            elem = elem[0] + elem[1].lower()
                        atomic_number = AtomicNum[elem]
                    except (IndexError, KeyError):
                        # Figure it out from the mass
                        atomic_number = AtomicNum[element_by_mass(mass)]
                    atype = AtomType(name=name, number=idx, mass=mass, atomic_number=atomic_number)
                    self.atom_types_str[atype.name] = atype
                    self.atom_types_int[atype.number] = atype
                    self.atom_types_tuple[(atype.name, atype.number)] = atype
                elif line[:4].upper() == 'DECL':
                    pass # Not really sure what this means
                elif line[:4].upper() == 'DEFA':
                    words = line.split()
                    if len(words) < 5:
                        warnings.warn('DEFA line has %d tokens; expected 5' % len(words))
                    else:
                        it = iter(words[1:5])
                        for tok, val in zip(it, it):
                            if val.upper() == 'NONE':
                                val = None
                            if tok.upper().startswith('FIRS'):
                                hpatch = val
                            elif tok.upper() == 'LAST':
                                tpatch = val
                            else:
                                warnings.warn('DEFA patch %s unknown' % val)
                elif line[:4].upper() in ('RESI', 'PRES'):
                    restype = line[:4].upper()
                    # Get the residue definition
                    words = line.split()
                    resname = words[1].upper()
                    if resname in self.residues:
                        warnings.warn('Replacing residue {}'.format(resname), ParameterWarning)
                    # Assign default patches
                    hpatches[resname] = hpatch
                    tpatches[resname] = tpatch
                    try:
                        charge = float(words[2])
                    except (IndexError, ValueError):
                        warnings.warn('No charge for %s' % resname)
                    if restype == 'RESI':
                        res = ResidueTemplate(resname)
                    elif restype == 'PRES':
                        res = PatchTemplate(resname)
                    else:
                        assert False, 'restype != RESI or PRES'
                    skip_adding_residue = False
                    line = next(f)
                    group = []
                    ictable = []
                    while line:
                        line = line.lstrip()
                        if line[:5].upper() == 'GROUP':
                            if group:
                                res.groups.append(group)
                            group = []
                        elif line[:4].upper() == 'ATOM':
                            words = line.split()
                            name = words[1].upper()
                            type = words[2].upper()
                            charge = float(words[3])
                            if 'ALPHA' in words:
                                # This is a polarizable atom.
                                alpha = float(words[words.index('ALPHA')+1])
                                thole = 1.3
                                drude_type = 'DRUD'
                                if 'THOLE' in words:
                                    thole = float(words[words.index('THOLE')+1])
                                if 'TYPE' in words:
                                    drude_type = words[words.index('TYPE')+1]
                                atom = DrudeAtom(name=name, type=type, charge=charge, alpha=alpha, thole=thole, drude_type=drude_type)
                            else:
                                atom = Atom(name=name, type=type, charge=charge)
                            group.append(atom)
                            res.add_atom(atom)
                        elif line[:6].upper() == 'DELETE':
                            words = line.split()
                            name = words[2].upper()
                            entity_type = words[1].upper()
                            if entity_type == 'ATOM':
                                res.delete_atoms.append(name)
                            elif entity_type == 'IMPR':
                                res.delete_impropers.append(words[2:5])
                            else:
                                warnings.warn('WARNING: Ignoring "%s" because entity type %s not '
                                              'used.' % (line.strip(), entity_type))
                        elif line.strip().upper() and line.split()[0].upper() in ('BOND', 'DOUBLE'):
                            it = iter([w.upper() for w in line.split()[1:]])
                            for a1, a2 in zip(it, it):
                                if restype == 'PRES':
                                    # Patches can have bonds that refer to atoms not in the patch, so store these in a list of tuples
                                    order = 1
                                    if line.split()[0].upper() == 'DOUBLE':
                                        order = 2
                                    res.add_bonds.append( (a1, a2, order) )
                                    continue

                                if a1.startswith('-'):
                                    res.head = res[a2]
                                    continue
                                if a2.startswith('-'):
                                    res.head = res[a1]
                                    continue
                                if a1.startswith('+'):
                                    res.tail = res[a2]
                                    continue
                                if a2.startswith('+'):
                                    res.tail = res[a1]
                                    continue
                                res.add_bond(a1, a2)
                        elif line[:4].upper() == 'CMAP':
                            pass
                        elif line[:5].upper() == 'DONOR':
                            pass
                        elif line[:6].upper() == 'ACCEPT':
                            pass
                        elif line[:8].upper() == 'LONEPAIR':
                            # See: https://www.charmm.org/charmm/documentation/by-version/c40b1/params/doc/lonepair/
                            # TODO: This currently doesn't handle some formats, like Note 3 in the above URL
                            words = line.split()
                            lptype_keyword = words[1][0:4].upper()
                            if not skip_adding_residue and lptype_keyword not in ['BISE', 'RELA']:
                                warnings.warn('LONEPAIR type %s not supported; only BISEctor and '
                                              'RELAtive supported' % words[1])
                                skip_adding_residue = True
                                break
                            a1, a2, a3, a4 = words[2:6]
                            keywords = {words[index][0:4].upper() : float(words[index+1])
                                        for index in range(6,len(words),2) }
                            r = keywords['DIST'] # angstrom
                            theta = keywords['ANGL'] # degrees
                            phi = keywords['DIHE'] # degrees
                            lptypes = { 'BISE' : 'bisector', 'RELA' : 'relative' }
                            lonepair = (lptypes[lptype_keyword], a1, a2, a3, a4, r, theta, phi) # TODO: Define a LonePair object?
                            res.lonepairs.append(lonepair)
                        elif line[:2].upper() == 'IC':
                            words = line.split()[1:]
                            ictable.append(([w.upper() for w in words[:4]],
                                            [float(w) for w in words[4:]]))
                        elif line[:3].upper() == 'END':
                            break
                        elif line[:5].upper() == 'PATCH':
                            it = iter(line.split()[1:])
                            for tok, val in zip(it, it):
                                if val.upper() == 'NONE': val = None
                                if tok.upper().startswith('FIRS'):
                                    hpatches[resname] = val
                                elif tok.upper().startswith('LAST'):
                                    tpatches[resname] = val
                        elif line[:4].upper() in ('IMPR', 'IMPH'):
                            it = iter(w.upper() for w in line.split()[1:])
                            for a1, a2, a3, a4 in zip(it, it, it, it):
                                res._impr.append((a1, a2, a3, a4))
                                if a2[0] == '-' or a3[0] == '-' or a4 == '-':
                                    res.head = res[a1]
                        elif line[:10].upper() == 'ANISOTROPY':
                            words = line.split()
                            atoms = [res[name] for name in words[1:5]]
                            keywords = {words[index].upper() : float(words[index+1])
                                        for index in range(5,len(words),2)}
                            a11 = float(keywords['A11'])
                            a22 = float(keywords['A22'])
                            atoms[0].anisotropy = DrudeAnisotropy(*atoms, a11=a11, a22=a22)
                        elif line[:4].upper() in ('RESI', 'PRES', 'MASS'):
                            # Back up a line and bail
                            break
                        line = next(f)
                    if group: res.groups.append(group)
                    _fit_IC_table(res, ictable)
                    if skip_adding_residue:
                        # Do not add this residue to the lookup library
                        continue
                    elif restype == 'RESI':
                        residues[resname] = res
                    elif restype == 'PRES':
                        patches[resname] = res
                    else:
                        assert False, 'restype != RESI or PRES'
                    # We parsed a line we need to look at. So don't update the
                    # iterator
                    continue
                # Get the next line and cycle through
                line = next(f)
                line_index += 1
        except StopIteration:
            pass

        # Go through the patches and add the appropriate one
        self.patches.update(patches)
        for resname, res in iteritems(residues):
            patch_name = hpatches[resname]
            if patch_name is not None:
                try:
                    res.first_patch = self.patches[patch_name]
                except KeyError:
                    warnings.warn('Patch %s not found' % patch_name)

            patch_name = tpatches[resname]
            if patch_name is not None:
                try:
                    res.last_patch = self.patches[patch_name]
                except KeyError:
                    warnings.warn('Patch %s not found' % patch_name)
        # Now update the residues and patches with the ones we parsed here
        self.residues.update(residues)

        if own_handle: f.close()

    def read_stream_file(self, sfile):
        """
        Reads RTF and PAR sections from a stream file and dispatches the
        sections to read_topology_file or read_parameter_file

        Parameters
        ----------
        sfile : str or CharmmStreamFile
            Stream file to parse
        """
        if isinstance(sfile, CharmmStreamFile):
            f = sfile
        else:
            f = CharmmStreamFile(sfile)

        title, section, comments = f.next_section()
        while title is not None and section is not None:
            words = title.lower().split()
            if words[1] == 'rtf':
                # This is a Residue Topology File section.
                self.read_topology_file(iter(section))
            elif words[1].startswith('para'):
                # This is a Parameter file section
                self.read_parameter_file(section, comments)
            title, section, comments = f.next_section()

    def write(self, top=None, par=None, str=None):
        """ Write a CHARMM parameter set to a file

        Parameters
        ----------
        top : str or file-like object, optional
            If provided, the atom types will be written to this file in RTF
            format.
        par : str or file-like object, optional
            If provided, the parameters will be written to this file in PAR
            format. Either this or the ``str`` argument *must* be provided
        str : str or file-like object, optional
            If provided, the atom types and parameters will be written to this
            file as separate RTF and PAR cards that can be read as a CHARMM
            stream file. Either this or the ``par`` argument *must* be provided

        Raises
        ------
        ValueError if both par and str are None
        """
        if par is None and str is None:
            raise ValueError('Must specify either par *or* str')

        if top is not None:
            if isinstance(top, string_types):
                f = genopen(top, 'w')
                ownhandle = True
            else:
                f = top
                ownhandle = False
            f.write('*>>>> CHARMM Topology file generated by ParmEd <<<<\n')
            f.write('*\n')
            self._write_top_to(f, True)
            if ownhandle: f.close()
        if par is not None:
            if isinstance(par, string_types):
                f = genopen(par, 'w')
                ownhandle = True
            else:
                f = par
                ownhandle = False
            f.write('*>>>> CHARMM Parameter file generated by ParmEd <<<<\n')
            f.write('*\n\n')
            self._write_par_to(f)
            if ownhandle: f.close()
        if str is not None:
            if isinstance(str, string_types):
                f = genopen(str, 'w')
                ownhandle = True
            else:
                f = str
                ownhandle = False
            self._write_str_to(f)
            if ownhandle: f.close()

    def _write_str_to(self, f):
        """ Private method to write stream items to open file object """
        f.write('read rtf card\n* Topology generated by ParmEd\n*\n')
        self._write_top_to(f, True)
        f.write('\nread para card\n* Parameters generated by ParmEd\n*\n')
        self._write_par_to(f)

    def _write_top_to(self, f, write_version):
        """ Private method to write topology items to open file object """
        if write_version:
            # This version is known to work
            f.write('36   1\n')
            f.write('\n')
        for i, (_, atom) in enumerate(iteritems(self.atom_types)):
            f.write('MASS %5d %-6s %9.5f\n' % (i+1, atom.name, atom.mass))
        if write_version:
            f.write('\nEND\n')

    def _write_par_to(self, f):
        """ Private method to write parameter items to open file object """
        # Find out what the 1-4 electrostatic scaling factors and the 1-4
        # van der Waals scaling factors are
        scee, scnb = set(), set()
        for _, typ in iteritems(self.dihedral_types):
            for t in typ:
                if t.scee: scee.add(t.scee)
                if t.scnb: scnb.add(t.scnb)
        if len(scee) > 1 or len(scnb) > 1:
            raise ValueError('Mixed 1-4 scaling not supported')
        scee = 1.0 if not scee else scee.pop()
        scnb = 1.0 if not scnb else scnb.pop()

        f.write('ATOMS\n')
        self._write_top_to(f, False)
        f.write('\nBONDS\n')
        written = set()
        for key, typ in iteritems(self.bond_types):
            if key in written: continue
            written.add(key); written.add(tuple(reversed(key)))
            f.write('%-6s %-6s %7.2f %10.4f\n' %
                    (key[0], key[1], typ.k, typ.req))
        f.write('\nANGLES\n')
        written = set()
        for key, typ in iteritems(self.angle_types):
            if key in written: continue
            written.add(key)
            written.add(tuple(reversed(key)))
            f.write('%-6s %-6s %-6s %7.2f %8.2f\n' % (key[0], key[1], key[2], typ.k, typ.theteq))
        f.write('\nDIHEDRALS\n')
        written = set()
        for key, typ in iteritems(self.dihedral_types):
            if key in written: continue
            written.add(key)
            written.add(tuple(reversed(key)))
            for tor in typ:
                f.write('%-6s %-6s %-6s %-6s %11.4f %2d %8.2f\n' %
                        (key[0], key[1], key[2], key[3], tor.phi_k, int(tor.per), tor.phase))
        f.write('\nIMPROPERS\n')
        written = set()
        for key, typ in sorted(iteritems(self.improper_periodic_types), key=lambda x: x[0]):
            f.write('%-6s %-6s %-6s %-6s %11.4f %2d %8.2f\n' %
                    (key[0], key[1], key[2], key[3], typ.phi_k, int(typ.per), typ.phase))
        for key, typ in iteritems(self.improper_types):
            f.write('%-6s %-6s %-6s %-6s %11.4f %2d %8.2f\n' %
                    (key[0], key[1], key[2], key[3], typ.psi_k, 0, typ.psi_eq))
        if self.cmap_types:
            f.write('\nCMAPS\n')
            written = set()
            for key, typ in iteritems(self.cmap_types):
                if key in written: continue
                written.add(key); written.add(tuple(reversed(key)))
                f.write('%-6s %-6s %-6s %-6s %-6s %-6s %-6s %-6s %5d\n\n' %
                        (key[0], key[1], key[2], key[3], key[4], key[5], key[6],
                         key[7], typ.resolution))
                i = 0
                for val in typ.grid:
                    if i:
                        if i % 5 == 0:
                            f.write('\n')
                            if i % typ.resolution == 0:
                                f.write('\n')
                                i = 0
                        elif i % typ.resolution == 0:
                            f.write('\n\n')
                            i = 0
                    i += 1
                    f.write(' %13.6f' % val)
                f.write('\n\n\n')
        f.write('\nNONBONDED  nbxmod  5 atom cdiel fshift vatom vdistance vfswitch -\ncutnb 14.0 '
                'ctofnb 12.0 ctonnb 10.0 eps 1.0 e14fac %s wmin 1.5%s\n\n' %
                (1/scee, ' GEOM' if self.combining_rule == 'geometric' else ''))
        for key, typ in iteritems(self.atom_types):
            f.write('%-6s %14.6f %10.6f %14.6f' % (key, 0.0, -abs(typ.epsilon), typ.rmin))
            if typ.epsilon == typ.epsilon_14 and typ.rmin == typ.rmin_14:
                f.write('%10.6f %10.6f %14.6f\n' % (0, -abs(typ.epsilon)/scnb, typ.rmin))
            else:
                f.write('%10.6f %10.6f %14.6f\n' % (0, -abs(typ.epsilon_14), typ.rmin_14))
        f.write('\nEND\n')

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def _fit_IC_table(res, ictable):
    """
    Determines cartesian coordinates from an internal coordinate table stored in
    CHARMM residue topology files

    Parameters
    ----------
    res : ResidueTemplate
        The residue template for which coordinates are being determined
    ictable : list[tuple(atoms, measurements)]
        The internal coordinate table

    Notes
    -----
    This method assigns an xx, xy, and xz attribute to ``res``. For the time
    being, this is just a placeholder, as its functionality has not yet been
    implemented (CHARMM does not use a 'traditional' Z-matrix, and I don't know
    of any existing code that will compute the proper cartesian coordinates from
    the form used by CHARMM)
    """
    for atom in res:
        atom.xx = atom.xy = atom.xz = 0.0



"""
This is a collection of all of the OpenMM functionality supported in ParmEd
"""

__all__ = ['StateDataReporter', 'NetCDFReporter', 'MdcrdReporter',
           'RestartReporter', 'ProgressReporter', 'EnergyMinimizerReporter',
           'utils', 'load_topology', 'XmlFile', 'energy_decomposition',
           'energy_decomposition_system', 'OpenMMParameterSet']

from parmed.openmm.reporters import (
        StateDataReporter, NetCDFReporter, MdcrdReporter, RestartReporter,
        ProgressReporter, EnergyMinimizerReporter,
)
from parmed.openmm.parameters import OpenMMParameterSet
from parmed.openmm.topsystem import load_topology
from parmed.openmm.utils import (energy_decomposition,
                                 energy_decomposition_system)
from parmed.openmm.xmlfile import XmlFile



""" A collection of utilities for use with Pandas objects """
from __future__ import absolute_import, division, print_function

import numpy as np
try:
    import pandas as pd
except ImportError:
    pd = None
from parmed.exceptions import ParameterWarning
from parmed.utils.six.moves import zip
from parmed.utils.six import iteritems
import warnings

# Utility function for generating a DataFrame based on a collection of items.
# The passed object must have an `atoms` attribute

def create_dataframe(obj):
    """ Creates a pandas.DataFrame object from the current instance

    Returns
    -------
    dataframe : :class:`pandas.DataFrame`

    Notes
    -----
    The DataFrame will be over all atoms. The columns will be the attributes
    of the atom (as well as its containing residue). Some columns will
    *always* exist. Others will only exist if those attributes have been set
    on the Atom instances (see the :class:`Atom` docs for possible
    attributes and their meaning). The columns that will always be present
    are:

        - number : int
        - name : str
        - type : str
        - atomic_number : int
        - charge : float
        - mass : float
        - nb_idx : int
        - solvent_radius : float
        - screen : float
        - occupancy : float
        - bfactor : float
        - altloc : str
        - tree : str
        - join : int
        - irotat : int
        - rmin : float
        - epsilon : float
        - rmin_14 : float
        - epsilon_14 : float
        - resname : str (name of the containing residue)
        - resid : int (Sequential index of the containing residue)
        - resnum : int (original residue number in the input structure)
        - chain : str (chain ID that the containing residue belongs to)

    The following attributes are optionally present if they were present in
    the original file defining the structure:

        - xx : float (x-coordinate position)
        - xy : float (y-coordinate position)
        - xz : float (z-coordinate position)
        - vx : float (x-coordinate velocity)
        - vy : float (y-coordinate velocity)
        - vz : float (z-coordinate velocity)
        - type_idx : int (integer type index for AMOEBA)
        - class_idx : int (integer class type index for AMOEBA)
        - multipole_111 : float (Monopole)
        - multipole_211 : float (1,1 Dipole component)
        - multipole_212 : float (1,2 Dipole component)
        - multipole_222 : float (2,2 Dipole component)
        - multipole_411 : float (1,1 Quadrupole component)
        - multipole_412 : float (1,2 Quadrupole component)
        - multipole_422 : float (2,2 Quadrupole component)
        - multipole_413 : float (1,3 Quadrupole component)
        - multipole_423 : float (2,3 Quadrupole component)
        - multipole_433 : float (3,3 Quadrupole component)
        - polarizability : float (dipole polarizability)
        - vdw_parent : int (index of the vdW parent atom of this atom)
        - U11 : float (U[1][1] of anisotropic b-factor tensor)
        - U22 : float (U[2][2] of anisotropic b-factor tensor)
        - U33 : float (U[3][3] of anisotropic b-factor tensor)
        - U12 : float (U[1][2] of anisotropic b-factor tensor)
        - U13 : float (U[1][3] of anisotropic b-factor tensor)
        - U23 : float (U[2][3] of anisotropic b-factor tensor)
    """
    if pd is None:
        raise ImportError('pandas is not available; cannot create a pandas '
                          'DataFrame from this Structure')
    ret = pd.DataFrame()

    atoms = obj.atoms

    ret['number'] = [atom.number for atom in atoms]
    ret['name'] = [atom.name for atom in atoms]
    ret['type'] = [atom.type for atom in atoms]
    ret['atomic_number'] = [atom.atomic_number for atom in atoms]
    ret['charge'] = [atom.charge for atom in atoms]
    ret['mass'] = [atom.mass for atom in atoms]
    ret['nb_idx'] = [atom.nb_idx for atom in atoms]
    ret['solvent_radius'] = [atom.solvent_radius for atom in atoms]
    ret['screen'] = [atom.screen for atom in atoms]
    ret['occupancy'] = [atom.occupancy for atom in atoms]
    ret['bfactor'] = [atom.bfactor for atom in atoms]
    ret['altloc'] = [atom.altloc for atom in atoms]
    ret['tree'] = [atom.tree for atom in atoms]
    ret['join'] = [atom.join for atom in atoms]
    ret['irotat'] = [atom.irotat for atom in atoms]
    ret['rmin'] = [atom.rmin for atom in atoms]
    ret['epsilon'] = [atom.epsilon for atom in atoms]
    ret['rmin_14'] = [atom.rmin_14 for atom in atoms]
    ret['epsilon_14'] = [atom.epsilon_14 for atom in atoms]
    ret['resname'] = [atom.residue.name for atom in atoms]
    ret['resid'] = [atom.residue.idx for atom in atoms]
    ret['resnum'] = [atom.residue.number for atom in atoms]
    ret['chain'] = [atom.residue.chain for atom in atoms]
    ret['segid'] = [atom.residue.segid for atom in atoms]

    # Now for optional attributes
    # Coordinates
    try:
        coords = pd.DataFrame(
                [[atom.xx, atom.xy, atom.xz] for atom in atoms],
                columns=['xx', 'xy', 'xz']
        )
    except AttributeError:
        pass
    else:
        ret = ret.join(coords)
    # Velocities
    try:
        vels = pd.DataFrame(
                [[atom.vx, atom.vy, atom.vz] for atom in atoms],
                columns=['vx', 'vy', 'vz']
        )
    except AttributeError:
        pass
    else:
        ret = ret.join(vels)
    # AMOEBA LJ type
    try:
        ret['type_idx'] = [atom.type_idx for atom in atoms]
    except AttributeError:
        pass
    # AMOEBA class type
    try:
        ret['class_idx'] = [atom.class_idx for atom in atoms]
    except AttributeError:
        pass
    # AMOEBA multipoles
    try:
        multipoles = pd.DataFrame(
                [atom.multipoles for atom in atoms],
                columns=['multipole_111', 'multipole_211', 'multipole_212',
                         'multipole_222', 'multipole_411', 'multipole_412',
                         'multipole_422', 'multipole_413', 'multipole_423',
                         'multipole_433']
        )
    except AttributeError:
        pass
    else:
        ret = ret.join(multipoles)
    # AMOEBA polarizabilities
    try:
        ret['polarizability'] = [atom.polarizability for atom in atoms]
    except AttributeError:
        pass
    # AMOEBA vdw parent atom
    try:
        ret['vdw_parent'] = [atom.vdw_parent.idx for atom in atoms]
    except AttributeError:
        pass
    # anisotropic b-factors
    none6 = [None] * 6
    anisos = [atom.anisou for atom in atoms]
    for i, aniso in enumerate(anisos):
        if hasattr(aniso, 'tolist'):
            anisos[i] = aniso.tolist()
    all_nones = True
    for i, aniso in enumerate(anisos):
        if aniso is None:
            anisos[i] = none6
        elif all_nones:
            all_nones = False
    if not all_nones:
        ret = ret.join(
                pd.DataFrame(anisos,
                    columns=['U11', 'U22', 'U33', 'U12', 'U13', 'U23'])
        )
    return ret

def load_dataframe(obj, dataframe):
    """
    Loads a DataFrame into the current object, setting atomic properties based
    on the entries of the DataFrame.  Supported atomic properties are:

        - number : int
        - name : str
        - type : str
        - atomic_number : int
        - charge : float
        - mass : float
        - nb_idx : int
        - solvent_radius : float
        - screen : float
        - occupancy : float
        - bfactor : float
        - altloc : str
        - tree : str
        - join : int
        - irotat : int
        - rmin : float
        - epsilon : float
        - rmin_14 : float
        - epsilon_14 : float
        - xx : float (x-coordinate position)
        - xy : float (y-coordinate position)
        - xz : float (z-coordinate position)
        - vx : float (x-coordinate velocity)
        - vy : float (y-coordinate velocity)
        - vz : float (z-coordinate velocity)
        - type_idx : int (integer type index for AMOEBA)
        - class_idx : int (integer class type index for AMOEBA)
        - multipole_111 : float (Monopole)
        - multipole_211 : float (1,1 Dipole component)
        - multipole_212 : float (1,2 Dipole component)
        - multipole_222 : float (2,2 Dipole component)
        - multipole_411 : float (1,1 Quadrupole component)
        - multipole_412 : float (1,2 Quadrupole component)
        - multipole_422 : float (2,2 Quadrupole component)
        - multipole_413 : float (1,3 Quadrupole component)
        - multipole_423 : float (2,3 Quadrupole component)
        - multipole_433 : float (3,3 Quadrupole component)
        - polarizability : float (dipole polarizability)
        - vdw_parent : int (index of the vdW parent atom of this atom)
        - segid : segment ID (similar to chain, but for CHARMM)
        - U11 : float (U[1][1] of anisotropic b-factor tensor)
        - U22 : float (U[2][2] of anisotropic b-factor tensor)
        - U33 : float (U[3][3] of anisotropic b-factor tensor)
        - U12 : float (U[1][2] of anisotropic b-factor tensor)
        - U13 : float (U[1][3] of anisotropic b-factor tensor)
        - U23 : float (U[2][3] of anisotropic b-factor tensor)

    The resname, resid, chain, and resnum attributes are ignored. Other
    attributes emit a ParameterWarning
    """
    atoms = obj.atoms

    def set_attribute(attr, data):
        """ Set the attribute list from the data """
        if len(data) != len(atoms):
            raise ValueError('Data does not match length of atoms list')
        for atom, x in zip(atoms, data):
            setattr(atom, attr, x)
    def set_residue_attr(attr, data):
        if len(data) != len(atoms):
            raise ValueError('Data does not match length of atoms list')
        for atom, x in zip(atoms, data):
            setattr(atom.residue, attr, x)

    multipoles = [None for i in range(10)]
    anisous = [None for i in range(6)]
    for key, data in iteritems(dataframe):
        if key == 'number':
            set_attribute('number', data)
        elif key == 'name':
            set_attribute('name', data)
        elif key == 'type':
            set_attribute('type', data)
        elif key == 'atomic_number':
            set_attribute('atomic_number', data)
        elif key == 'charge':
            set_attribute('charge', data)
        elif key == 'mass':
            set_attribute('mass', data)
        elif key == 'nb_idx':
            set_attribute('nb_idx', data)
        elif key == 'solvent_radius':
            set_attribute('solvent_radius', data)
        elif key == 'screen':
            set_attribute('screen', data)
        elif key == 'occupancy':
            set_attribute('occupancy', data)
        elif key == 'bfactor':
            set_attribute('bfactor', data)
        elif key == 'altloc':
            set_attribute('altloc', data)
        elif key == 'tree':
            set_attribute('tree', data)
        elif key == 'join':
            set_attribute('join', data)
        elif key == 'irotat':
            set_attribute('irotat', data)
        elif key == 'rmin':
            set_attribute('rmin', data)
        elif key == 'epsilon':
            set_attribute('epsilon', data)
        elif key == 'rmin_14':
            set_attribute('rmin_14', data)
        elif key == 'epsilon_14':
            set_attribute('epsilon_14', data)
        elif key == 'xx':
            set_attribute('xx', data)
        elif key == 'xy':
            set_attribute('xy', data)
        elif key == 'xz':
            set_attribute('xz', data)
        elif key == 'vx':
            set_attribute('vx', data)
        elif key == 'vy':
            set_attribute('vy', data)
        elif key == 'vz':
            set_attribute('vz', data)
        elif key == 'type_idx':
            set_attribute('type_idx', data)
        elif key == 'class_idx':
            set_attribute('class_idx', data)
        elif key == 'vdw_parent':
            if len(data) != len(atoms):
                raise ValueError('vdw_parent length not equal to natom')
            for atom, parent in zip(atoms, data):
                atom.vdw_parent = atoms[parent]
        elif key == 'polarizability':
            set_attribute('polarizability', data)
        elif key == 'multipole_111':
            multipoles[0] = data
        elif key == 'multipole_211':
            multipoles[1] = data
        elif key == 'multipole_212':
            multipoles[2] = data
        elif key == 'multipole_213':
            multipoles[3] = data
        elif key == 'multipole_411':
            multipoles[4] = data
        elif key == 'multipole_412':
            multipoles[5] = data
        elif key == 'multipole_422':
            multipoles[6] = data
        elif key == 'multipole_413':
            multipoles[7] = data
        elif key == 'multipole_423':
            multipoles[8] = data
        elif key == 'multipole_433':
            multipoles[9] = data
        elif key == 'U11':
            anisous[0] = data
        elif key == 'U22':
            anisous[1] = data
        elif key == 'U33':
            anisous[2] = data
        elif key == 'U12':
            anisous[3] = data
        elif key == 'U13':
            anisous[4] = data
        elif key == 'U23':
            anisous[5] = data
        elif key in ('resname', 'resid', 'resnum', 'chain', 'segid'):
            set_residue_attr(key, data)
            continue
        else:
            warnings.warn('Atomic property %s not recognized' % key,
                          ParameterWarning)

    # Now combine the multipoles and anisous if they are all specified
    if not None in anisous:
        set_attribute('anisou', np.vstack(anisous).T)
    if not None in multipoles:
        set_attribute('multipoles', np.vstack(multipoles).T)



"""
This module contains basic information and functionality related to individual
residues in typical biopolymers.
"""

__all__ = ['AminoAcidResidue', 'RNAResidue', 'DNAResidue', 'ALA', 'ARG', 'ASN',
           'ASP', 'CYS', 'GLU', 'GLN', 'GLY', 'HIS', 'HYP', 'ILE', 'LEU', 'LYS',
           'MET', 'PHE', 'PRO', 'SER', 'THR', 'TRP', 'TYR', 'VAL', 'DA', 'DT',
           'DG', 'DC', 'A', 'U', 'G', 'C', 'SOLVENT_NAMES', 'EXTRA_POINT_NAMES',
           'CATION_NAMES', 'ANION_NAMES', 'ALLION_NAMES']

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class BiomolecularResidue(object):
    """ Base class for different classes of biopolymer residues """
    _all_residues_by_name = dict()
    _all_residues_by_abbr = dict()
    _all_residues_by_symbol = dict()
    all_residues = []

    def __init_(self, *args, **kwargs):
        raise NotImplementedError('BiomolecularResidue must be subclassed')

    @classmethod
    def get(cls, key):
        raise NotImplementedError('BiomolecularResidue must be subclassed')

    def __str__(self):
        return self.name

    @classmethod
    def has(cls, thing):
        """
        Determines if a particular BiomolecularResidue or residue name is
        present in this classification of biomolecular residues

        Parameters
        ----------
        thing : str or :class:`BiomolecularResidue`

        Returns
        -------
        contains : bool
            If the residue or residue name *is* of this type, True. Otherwise,
            False.
        """
        if isinstance(thing, BiomolecularResidue):
            return thing in cls.all_residues
        try:
            cls.get(thing)
        except KeyError:
            return False
        else:
            return True

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class AminoAcidResidue(BiomolecularResidue):
    """
    An individual amino acid residue.

    Parameters
    ----------
    name : str
        The name of the residue
    abbr : str
        The 3-letter abbreviation of the amino acid residue
    symbol : str
        The 1-letter symbol of the amino acid
    aliases : list of str, optional
        A list of other abbreviations that *also* refer to this residue

    Raises
    ------
    ValueError
        If any aliases have the same abbreviation as *other*
    """
    _all_residues_by_name = dict()
    _all_residues_by_abbr = dict()
    _all_residues_by_symbol = dict()
    all_residues = []

    def __init__(self, name, abbr, symbol, aliases=None):
        self.name = name
        self.abbr = abbr
        self.symbol = symbol
        type(self)._all_residues_by_name[name.upper()] = self
        type(self)._all_residues_by_abbr[abbr.upper()] = self
        if symbol is not None:
            type(self)._all_residues_by_symbol[symbol.upper()] = self
        type(self).all_residues.append(self)
        if aliases is not None:
            for alias in aliases:
                alias = alias.upper()
                if alias in type(self)._all_residues_by_abbr:
                    raise ValueError('%s is already an abbreviation' % alias)
                type(self)._all_residues_by_abbr[alias] = self

    def __repr__(self):
        return '<Amino Acid Residue %s: %s [%s]>' % (self.name, self.abbr,
                self.symbol)

    @classmethod
    def get(cls, key, abbronly=False):
        """
        Gets the amino acid corresponding to either the residue name, 3-letter
        abbreviation or 1-letter symbol. It is case-insensitive.

        Parameters
        ----------
        key : str
            1-letter symbol, 3-letter abbreviation, or residue name
        abbronly : bool
            If True, only look for the 3-letter abbreviation (not the 1-letter
            symbol)

        Returns
        -------
        residue : :class:`AminoAcidResidue`
            The residue corresponding to the given key

        Raises
        ------
        KeyError if ``key`` is not a symbol, abbreviation, or case-insensitive
        name of an amino acid residue, or any of its abbreviations.
        """
        if len(key) == 1 and not abbronly:
            return cls._all_residues_by_symbol[key.upper()]
        if len(key) == 3:
            return cls._all_residues_by_abbr[key.upper()]
        # Handle C- and N-termini that may be prepended with C or N
        if len(key) == 4 and key[0].upper() in 'CN':
            return cls._all_residues_by_abbr[key[1:].upper()]
        return cls._all_residues_by_name[key.upper()]

ALA = AminoAcidResidue('Alanine', 'ALA', 'A')
ARG = AminoAcidResidue('Arginine', 'ARG', 'R')
ASN = AminoAcidResidue('Asparagine', 'ASN', 'N')
ASP = AminoAcidResidue('Aspartate' ,'ASP', 'D', ['ASH', 'AS4'])
CYS = AminoAcidResidue('Cysteine', 'CYS', 'C', ['CYM', 'CYX'])
GLU = AminoAcidResidue('Glutamate', 'GLU', 'E', ['GLH', 'GL4'])
GLN = AminoAcidResidue('Glutamine', 'GLN', 'Q')
GLY = AminoAcidResidue('Glycine', 'GLY', 'G')
HIS = AminoAcidResidue('Histidine', 'HIS', 'H', ['HIP', 'HIE', 'HID'])
HYP = AminoAcidResidue('Hydroxyproline', 'HYP', None)
ILE = AminoAcidResidue('Isoleucine', 'ILE', 'I')
LEU = AminoAcidResidue('Leucine', 'LEU', 'L')
LYS = AminoAcidResidue('Lysine', 'LYS', 'K', ['LYN'])
MET = AminoAcidResidue('Methionine', 'MET', 'M')
PHE = AminoAcidResidue('Phenylalanine', 'PHE', 'F')
PRO = AminoAcidResidue('Proline', 'PRO', 'P')
SER = AminoAcidResidue('Serine', 'SER', 'S')
THR = AminoAcidResidue('Threonine', 'THR', 'T')
TRP = AminoAcidResidue('Tryptophan', 'TRP', 'W')
TYR = AminoAcidResidue('Tyrosine', 'TYR', 'Y')
VAL = AminoAcidResidue('Valine', 'VAL', 'V')

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class DNAResidue(BiomolecularResidue):
    """ An individual DNA residue

    Parameters
    ----------
    name : str
        The name of the residue
    abbr : str
        The abbreviation of the nucleic acid residue
    aliases : list of str, optional
        A list of other abbreviations that *also* refer to this residue
    """
    _all_residues_by_name = dict()
    _all_residues_by_abbr = dict()
    _all_residues_by_symbol = dict()
    all_residues = []

    def __init__(self, name, abbr, aliases=None):
        self.name = name
        self.abbr = abbr
        type(self)._all_residues_by_name[name.upper()] = self
        type(self)._all_residues_by_abbr[abbr.upper()] = self
        type(self).all_residues.append(self)
        if aliases is not None:
            for alias in aliases:
                alias = alias.upper()
                if alias in type(self)._all_residues_by_abbr:
                    raise ValueError('%s is already an abbreviation' % alias)
                type(self)._all_residues_by_abbr[alias] = self

    def __repr__(self):
        return '<DNA Residue %s: %s>' % (self.name, self.abbr)

    @classmethod
    def get(cls, key):
        """
        Gets the nucleic acid corresponding to either the residue name or
        abbreviation. It is case-insensitive.

        Parameters
        ----------
        key : str
            abbreviation or residue name

        Returns
        -------
        residue : :class:`DNAResidue`
            The residue corresponding to the given key

        Raises
        ------
        KeyError if ``key`` is not a recognized residue name or abbreviation for
        an DNA residue.
        """
        try:
            if key[-1] in '35':
                return cls._all_residues_by_abbr[key[:-1].upper()]
            return cls._all_residues_by_abbr[key.upper()]
        except KeyError:
            return cls._all_residues_by_name[key.upper()]

class RNAResidue(DNAResidue):
    """ An individual RNA residue

    Parameters
    ----------
    name : str
        The name of the residue
    abbr : str
        The abbreviation of the nucleic acid residue
    aliases : list of str, optional
        A list of other abbreviations that *also* refer to this residue
    """
    _all_residues_by_name = dict()
    _all_residues_by_abbr = dict()
    _all_residues_by_symbol = dict()
    all_residues = []

    def __repr__(self):
        return '<RNA Residue %s: %s>' % (self.name, self.abbr)

    @classmethod
    def get(cls, key):
        """
        Gets the nucleic acid corresponding to either the residue name or
        abbreviation. It is case-insensitive.

        Parameters
        ----------
        key : str
            abbreviation or residue name

        Returns
        -------
        residue : :class:`RNAResidue`
            The residue corresponding to the given key

        Raises
        ------
        KeyError if ``key`` is not a recognized residue name or abbreviation for
        an RNA residue.
        """
        try:
            if key[-1] in '35':
                return cls._all_residues_by_abbr[key[:-1].upper()]
            return cls._all_residues_by_abbr[key.upper()]
        except KeyError:
            return cls._all_residues_by_name[key.upper()]

DG = DNAResidue('Guanine', 'DG', ['GUA', 'DG5', 'DG3', 'DGN'])
DC = DNAResidue('Cytosine', 'DC', ['CYT', 'DC5', 'DC3', 'DCN', 'DCP'])
DA = DNAResidue('Adenine', 'DA', ['ADE', 'DA5', 'DA3', 'DAN', 'DAP'])
DT = DNAResidue('Thymine', 'DT', ['THY', 'DT5', 'DT3'])
G = RNAResidue('Guanine', 'G', ['GUA', 'G5', 'G3', 'GN', 'RG', 'RG3', 'RG5', 'RGN'])
C = RNAResidue('Cytosine', 'C', ['CYT', 'CP', 'C5', 'C3', 'CN', 'RC', 'RC5', 'RC3', 'RCN'])
A = RNAResidue('Adenine', 'A', ['ADE', 'AP', 'A5', 'A3', 'AN', 'RA', 'RA3', 'RA5'])
U = RNAResidue('Uracil', 'U', ['URA', 'U3', 'U5', 'UN', 'RU', 'RU3', 'RU5', 'RUN'])
T = RNAResidue('Thymine', 'T', ['THY', 'T3', 'T5', 'TN', 'RT', 'RT3', 'RT5', 'RTN'])

WATER_NAMES = {'WAT', 'HOH', 'TIP3', 'TIP4', 'TIP5', 'SPCE', 'SPC'}
SOLVENT_NAMES = WATER_NAMES | {'SOL'}
EXTRA_POINT_NAMES = {'EP', 'LP'}
CATION_NAMES = {'Na+', 'Li+', 'Mg+', 'Rb+', 'MG', 'Cs+', 'POT', 'SOD', 'MG2',
                'CAL', 'RUB', 'LIT', 'ZN2', 'CD2', 'NA', 'K+', 'K', 'NA+'}
ANION_NAMES = {'Cl-', 'Br-', 'F-', 'I-', 'CLA', 'CL', 'BR', 'CL-'}
ALLION_NAMES = CATION_NAMES | ANION_NAMES
